"""
Copied from: https://github.com/tensorflow/models/blob/master/official/vision/image_classification/augment.py
Midified according to: https://github.com/rwightman/pytorch-image-models/blob/master/timm/data/auto_augment.py
"""
# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""AutoAugment and RandAugment policies for enhanced image preprocessing.

AutoAugment Reference: https://arxiv.org/abs/1805.09501
RandAugment Reference: https://arxiv.org/abs/1909.13719
"""

import math
import tensorflow as tf
from typing import Any, Dict, List, Optional, Text, Tuple, Union

# from tensorflow.python.keras.layers.preprocessing import image_preprocessing as image_ops
from keras.layers.preprocessing import image_preprocessing as image_ops

# This signifies the max integer that the controller RNN could predict for the
# augmentation scheme.
_MAX_LEVEL = 10.0


def to_4d(image: tf.Tensor) -> tf.Tensor:
    """Converts an input Tensor to 4 dimensions.

    4D image => [N, H, W, C] or [N, C, H, W]
    3D image => [1, H, W, C] or [1, C, H, W]
    2D image => [1, H, W, 1]

    Args:
      image: The 2/3/4D input tensor.

    Returns:
      A 4D image tensor.

    Raises:
      `TypeError` if `image` is not a 2/3/4D tensor.

    """
    shape = tf.shape(image)
    original_rank = tf.rank(image)
    left_pad = tf.cast(tf.less_equal(original_rank, 3), dtype=tf.int32)
    right_pad = tf.cast(tf.equal(original_rank, 2), dtype=tf.int32)
    new_shape = tf.concat(
        [
            tf.ones(shape=left_pad, dtype=tf.int32),
            shape,
            tf.ones(shape=right_pad, dtype=tf.int32),
        ],
        axis=0,
    )
    return tf.reshape(image, new_shape)


def from_4d(image: tf.Tensor, ndims: tf.Tensor) -> tf.Tensor:
    """Converts a 4D image back to `ndims` rank."""
    shape = tf.shape(image)
    begin = tf.cast(tf.less_equal(ndims, 3), dtype=tf.int32)
    end = 4 - tf.cast(tf.equal(ndims, 2), dtype=tf.int32)
    new_shape = shape[begin:end]
    return tf.reshape(image, new_shape)


def _convert_translation_to_transform(translations: tf.Tensor) -> tf.Tensor:
    """Converts translations to a projective transform.

    The translation matrix looks like this:
      [[1 0 -dx]
       [0 1 -dy]
       [0 0 1]]

    Args:
      translations: The 2-element list representing [dx, dy], or a matrix of
        2-element lists representing [dx dy] to translate for each image. The
        shape must be static.

    Returns:
      The transformation matrix of shape (num_images, 8).

    Raises:
      `TypeError` if
        - the shape of `translations` is not known or
        - the shape of `translations` is not rank 1 or 2.

    """
    translations = tf.convert_to_tensor(translations, dtype=tf.float32)
    if translations.get_shape().ndims is None:
        raise TypeError("translations rank must be statically known")
    elif len(translations.get_shape()) == 1:
        translations = translations[None]
    elif len(translations.get_shape()) != 2:
        raise TypeError("translations should have rank 1 or 2.")
    num_translations = tf.shape(translations)[0]

    return tf.concat(
        values=[
            tf.ones((num_translations, 1), tf.dtypes.float32),
            tf.zeros((num_translations, 1), tf.dtypes.float32),
            -translations[:, 0, None],
            tf.zeros((num_translations, 1), tf.dtypes.float32),
            tf.ones((num_translations, 1), tf.dtypes.float32),
            -translations[:, 1, None],
            tf.zeros((num_translations, 2), tf.dtypes.float32),
        ],
        axis=1,
    )


def _convert_angles_to_transform(angles: tf.Tensor, image_width: tf.Tensor, image_height: tf.Tensor) -> tf.Tensor:
    """Converts an angle or angles to a projective transform.

    Args:
      angles: A scalar to rotate all images, or a vector to rotate a batch of
        images. This must be a scalar.
      image_width: The width of the image(s) to be transformed.
      image_height: The height of the image(s) to be transformed.

    Returns:
      A tensor of shape (num_images, 8).

    Raises:
      `TypeError` if `angles` is not rank 0 or 1.

    """
    angles = tf.convert_to_tensor(angles, dtype=tf.float32)
    if len(angles.get_shape()) == 0:  # pylint:disable=g-explicit-length-test
        angles = angles[None]
    elif len(angles.get_shape()) != 1:
        raise TypeError("Angles should have a rank 0 or 1.")
    x_offset = ((image_width - 1) - (tf.math.cos(angles) * (image_width - 1) - tf.math.sin(angles) * (image_height - 1))) / 2.0
    y_offset = ((image_height - 1) - (tf.math.sin(angles) * (image_width - 1) + tf.math.cos(angles) * (image_height - 1))) / 2.0
    num_angles = tf.shape(angles)[0]
    return tf.concat(
        values=[
            tf.math.cos(angles)[:, None],
            -tf.math.sin(angles)[:, None],
            x_offset[:, None],
            tf.math.sin(angles)[:, None],
            tf.math.cos(angles)[:, None],
            y_offset[:, None],
            tf.zeros((num_angles, 2), tf.dtypes.float32),
        ],
        axis=1,
    )


def marge_two_transforms(aa: tf.Tensor, bb: tf.Tensor) -> tf.Tensor:
    cc = tf.reshape(tf.concat([aa, [[1.0]]], axis=1), [3, 3])
    dd = tf.reshape(tf.concat([bb, [[1.0]]], axis=1), [3, 3])
    return tf.reshape(tf.matmul(cc, dd), [1, -1])[:, :8]


def transform(image: tf.Tensor, transforms) -> tf.Tensor:
    """Prepares input data for `image_ops.transform`."""
    original_ndims = tf.rank(image)
    transforms = tf.convert_to_tensor(transforms, dtype=tf.float32)
    if transforms.shape.rank == 1:
        transforms = transforms[None]
    image = to_4d(image)
    # image = image_ops.transform(images=image, transforms=transforms, interpolation="nearest", fill_mode="constant")
    image = image_ops.transform(images=image, transforms=transforms, interpolation="bilinear", fill_mode="constant")
    return from_4d(image, original_ndims)


def blend(image1: tf.Tensor, image2: tf.Tensor, factor: float) -> tf.Tensor:
    """Blend image1 and image2 using 'factor'.

    Factor can be above 0.0.  A value of 0.0 means only image1 is used.
    A value of 1.0 means only image2 is used.  A value between 0.0 and
    1.0 means we linearly interpolate the pixel values between the two
    images.  A value greater than 1.0 "extrapolates" the difference
    between the two pixel values, and we clip the results to values
    between 0 and 255.

    Args:
      image1: An image Tensor of type uint8.
      image2: An image Tensor of type uint8.
      factor: A floating point value above 0.0.

    Returns:
      A blended image Tensor of type uint8.
    """
    if factor == 0.0:
        return tf.convert_to_tensor(image1)
    if factor == 1.0:
        return tf.convert_to_tensor(image2)

    image1 = tf.cast(image1, tf.float32)
    image2 = tf.cast(image2, tf.float32)

    difference = image2 - image1
    scaled = factor * difference

    # Do addition in float.
    temp = tf.cast(image1, tf.float32) + scaled

    # Interpolate
    if factor > 0.0 and factor < 1.0:
        # Interpolation means we always stay within 0 and 255.
        return tf.cast(temp, tf.uint8)

    # Extrapolate:
    #
    # We need to clip and then cast.
    return tf.cast(tf.clip_by_value(temp, 0.0, 255.0), tf.uint8)


def cutout(image: tf.Tensor, pad_size: int, replace: int = 0) -> tf.Tensor:
    """Apply cutout (https://arxiv.org/abs/1708.04552) to image.

    This operation applies a (2*pad_size x 2*pad_size) mask of zeros to
    a random location within `img`. The pixel values filled in will be of the
    value `replace`. The located where the mask will be applied is randomly
    chosen uniformly over the whole image.

    Args:
      image: An image Tensor of type uint8.
      pad_size: Specifies how big the zero mask that will be generated is that is
        applied to the image. The mask will be of size (2*pad_size x 2*pad_size).
      replace: What pixel value to fill in the image in the area that has the
        cutout mask applied to it.

    Returns:
      An image Tensor that is of type uint8.
    """
    image_height = tf.shape(image)[0]
    image_width = tf.shape(image)[1]

    # Sample the center location in the image where the zero mask will be applied.
    cutout_center_height = tf.random.uniform(shape=[], minval=0, maxval=image_height, dtype=tf.int32)

    cutout_center_width = tf.random.uniform(shape=[], minval=0, maxval=image_width, dtype=tf.int32)

    lower_pad = tf.maximum(0, cutout_center_height - pad_size)
    upper_pad = tf.maximum(0, image_height - cutout_center_height - pad_size)
    left_pad = tf.maximum(0, cutout_center_width - pad_size)
    right_pad = tf.maximum(0, image_width - cutout_center_width - pad_size)

    cutout_shape = [image_height - (lower_pad + upper_pad), image_width - (left_pad + right_pad)]
    padding_dims = [[lower_pad, upper_pad], [left_pad, right_pad]]
    mask = tf.pad(tf.zeros(cutout_shape, dtype=image.dtype), padding_dims, constant_values=1)
    mask = tf.expand_dims(mask, -1)
    mask = tf.tile(mask, [1, 1, 3])
    image = tf.where(tf.equal(mask, 0), tf.ones_like(image, dtype=image.dtype) * replace, image)
    return image


def solarize(image: tf.Tensor, threshold: int = 128) -> tf.Tensor:
    # For each pixel in the image, select the pixel
    # if the value is less than the threshold.
    # Otherwise, subtract 255 from the pixel.
    threshold = tf.cast(threshold, image.dtype)
    return tf.where(image < threshold, image, 255 - image)


def solarize_add(image: tf.Tensor, addition: int = 0, threshold: int = 128) -> tf.Tensor:
    # For each pixel in the image less than threshold
    # we add 'addition' amount to it and then clip the
    # pixel value to be between 0 and 255. The value
    # of 'addition' is between -128 and 128.
    added_image = tf.cast(image, tf.int32) + addition
    added_image = tf.cast(tf.clip_by_value(added_image, 0, 255), tf.uint8)
    threshold = tf.cast(threshold, image.dtype)
    return tf.where(image < threshold, added_image, image)


def color(image: tf.Tensor, factor: float) -> tf.Tensor:
    """Equivalent of PIL Color."""
    degenerate = tf.image.grayscale_to_rgb(tf.image.rgb_to_grayscale(image))
    return blend(degenerate, image, factor)


def contrast(image: tf.Tensor, factor: float) -> tf.Tensor:
    """Equivalent of PIL Contrast."""
    degenerate = tf.image.rgb_to_grayscale(image)
    # Cast before calling tf.histogram.
    degenerate = tf.cast(degenerate, tf.int32)

    # Compute the grayscale histogram, then compute the mean pixel value,
    # and create a constant image size of that value.  Use that as the
    # blending degenerate target of the original image.
    hist = tf.histogram_fixed_width(degenerate, [0, 255], nbins=256)
    mean = tf.reduce_sum(tf.cast(hist, tf.float32)) / 256.0
    degenerate = tf.ones_like(degenerate, dtype=tf.float32) * mean
    degenerate = tf.clip_by_value(degenerate, 0.0, 255.0)
    degenerate = tf.image.grayscale_to_rgb(tf.cast(degenerate, tf.uint8))
    return blend(degenerate, image, factor)


def brightness(image: tf.Tensor, factor: float) -> tf.Tensor:
    """Equivalent of PIL Brightness."""
    degenerate = tf.zeros_like(image)
    return blend(degenerate, image, factor)


def posterize(image: tf.Tensor, bits: int) -> tf.Tensor:
    """Equivalent of PIL Posterize."""
    shift = tf.cast(8 - bits, image.dtype)
    return tf.bitwise.left_shift(tf.bitwise.right_shift(image, shift), shift)


def rotate(image: tf.Tensor, degrees: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Rotates the image by degrees either clockwise or counterclockwise.

    Args:
      image: An image Tensor of type uint8.
      degrees: Float, a scalar angle in degrees to rotate all images by. If
        degrees is positive the image will be rotated clockwise otherwise it will
        be rotated counterclockwise.

    Returns:
      The rotated version of image.

    """
    # Convert from degrees to radians.
    degrees_to_radians = math.pi / 180.0
    radians = tf.cast(degrees * degrees_to_radians, tf.float32)

    image_height = tf.cast(tf.shape(image)[0], tf.float32)
    image_width = tf.cast(tf.shape(image)[1], tf.float32)
    transforms = _convert_angles_to_transform(angles=radians, image_width=image_width, image_height=image_height)
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def translate_x_relative(image: tf.Tensor, level: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL Translate in X dimension."""
    pixels = level * image.shape[0]
    transforms = _convert_translation_to_transform([-pixels, 0])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def translate_y_relative(image: tf.Tensor, level: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL Translate in Y dimension."""
    pixels = level * image.shape[1]
    transforms = _convert_translation_to_transform([0, -pixels])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def translate_x(image: tf.Tensor, pixels: int, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL Translate in X dimension."""
    transforms = _convert_translation_to_transform([-pixels, 0])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def translate_y(image: tf.Tensor, pixels: int, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL Translate in Y dimension."""
    transforms = _convert_translation_to_transform([0, -pixels])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def shear_x(image: tf.Tensor, level: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL Shearing in X dimension."""
    # Shear parallel to x axis is a projective transform
    # with a matrix form of:
    # [1  level
    #  0  1].
    transforms = tf.convert_to_tensor([[1.0, level, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def shear_y(image: tf.Tensor, level: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL Shearing in Y dimension."""
    # Shear parallel to y axis is a projective transform
    # with a matrix form of:
    # [1  0
    #  level  1].
    transforms = tf.convert_to_tensor([[1.0, 0.0, 0.0, level, 1.0, 0.0, 0.0, 0.0]])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def scale_x(image: tf.Tensor, level: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL scaling in X dimension."""
    transforms = tf.convert_to_tensor([[level, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def scale_y(image: tf.Tensor, level: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL scaling in Y dimension."""
    transforms = tf.convert_to_tensor([[1.0, 0.0, 0.0, 0.0, level, 0.0, 0.0, 0.0]])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def scale(image: tf.Tensor, level: float, replace: int, return_affine_matrix: bool = False) -> tf.Tensor:
    """Equivalent of PIL scaling in X and Y dimension."""
    transforms = tf.convert_to_tensor([[level, 0.0, 0.0, 0.0, level, 0.0, 0.0, 0.0]])
    if return_affine_matrix:
        return transforms

    image = transform(image=wrap(image), transforms=transforms)
    image = unwrap(image, replace)
    return image


def autocontrast(image: tf.Tensor) -> tf.Tensor:
    """Implements Autocontrast function from PIL using TF ops.

    Args:
      image: A 3D uint8 tensor.

    Returns:
      The image after it has had autocontrast applied to it and will be of type
      uint8.
    """

    def scale_channel(image: tf.Tensor) -> tf.Tensor:
        """Scale the 2D image using the autocontrast rule."""
        # A possibly cheaper version can be done using cumsum/unique_with_counts
        # over the histogram values, rather than iterating over the entire image.
        # to compute mins and maxes.
        lo = tf.cast(tf.reduce_min(image), tf.float32)
        hi = tf.cast(tf.reduce_max(image), tf.float32)

        # Scale the image, making the lowest value 0 and the highest value 255.
        def scale_values(im):
            scale = 255.0 / (hi - lo)
            offset = -lo * scale
            im = tf.cast(im, tf.float32) * scale + offset
            im = tf.clip_by_value(im, 0.0, 255.0)
            return tf.cast(im, tf.uint8)

        result = tf.cond(hi > lo, lambda: scale_values(image), lambda: image)
        return result

    # Assumes RGB for now.  Scales each channel independently
    # and then stacks the result.
    s1 = scale_channel(image[:, :, 0])
    s2 = scale_channel(image[:, :, 1])
    s3 = scale_channel(image[:, :, 2])
    image = tf.stack([s1, s2, s3], 2)
    return image


def sharpness(image: tf.Tensor, factor: float) -> tf.Tensor:
    """Implements Sharpness function from PIL using TF ops."""
    orig_image = image
    image = tf.cast(image, tf.float32)
    # Make image 4D for conv operation.
    image = tf.expand_dims(image, 0)
    # SMOOTH PIL Kernel.
    kernel = tf.constant([[1, 1, 1], [1, 5, 1], [1, 1, 1]], dtype=tf.float32, shape=[3, 3, 1, 1]) / 13.0
    # Tile across channel dimension.
    kernel = tf.tile(kernel, [1, 1, 3, 1])
    strides = [1, 1, 1, 1]
    degenerate = tf.nn.depthwise_conv2d(image, kernel, strides, padding="VALID", dilations=[1, 1])
    degenerate = tf.clip_by_value(degenerate, 0.0, 255.0)
    degenerate = tf.squeeze(tf.cast(degenerate, tf.uint8), [0])

    # For the borders of the resulting image, fill in the values of the
    # original image.
    mask = tf.ones_like(degenerate)
    padded_mask = tf.pad(mask, [[1, 1], [1, 1], [0, 0]])
    padded_degenerate = tf.pad(degenerate, [[1, 1], [1, 1], [0, 0]])
    result = tf.where(tf.equal(padded_mask, 1), padded_degenerate, orig_image)

    # Blend the final result.
    return blend(result, orig_image, factor)


def equalize(image: tf.Tensor) -> tf.Tensor:
    """Implements Equalize function from PIL using TF ops."""

    def scale_channel(im, c):
        """Scale the data in the channel to implement equalize."""
        im = tf.cast(im[:, :, c], tf.int32)
        # Compute the histogram of the image channel.
        histo = tf.histogram_fixed_width(im, [0, 255], nbins=256)

        # For the purposes of computing the step, filter out the nonzeros.
        nonzero = tf.where(tf.not_equal(histo, 0))
        nonzero_histo = tf.reshape(tf.gather(histo, nonzero), [-1])
        step = (tf.reduce_sum(nonzero_histo) - nonzero_histo[-1]) // 255

        def build_lut(histo, step):
            # Compute the cumulative sum, shifting by step // 2
            # and then normalization by step.
            lut = (tf.cumsum(histo) + (step // 2)) // step
            # Shift lut, prepending with 0.
            lut = tf.concat([[0], lut[:-1]], 0)
            # Clip the counts to be in range.  This is done
            # in the C code for image.point.
            return tf.clip_by_value(lut, 0, 255)

        # If step is zero, return the original image.  Otherwise, build
        # lut from the full histogram and step and then index from it.
        result = tf.cond(tf.equal(step, 0), lambda: im, lambda: tf.gather(build_lut(histo, step), im))

        return tf.cast(result, tf.uint8)

    # Assumes RGB for now.  Scales each channel independently
    # and then stacks the result.
    s1 = scale_channel(image, 0)
    s2 = scale_channel(image, 1)
    s3 = scale_channel(image, 2)
    image = tf.stack([s1, s2, s3], 2)
    return image


def invert(image: tf.Tensor) -> tf.Tensor:
    """Inverts the image pixels."""
    image = tf.convert_to_tensor(image)
    return 255 - image


def wrap(image: tf.Tensor) -> tf.Tensor:
    """Returns 'image' with an extra channel set to all 1s."""
    shape = tf.shape(image)
    extended_channel = tf.ones([shape[0], shape[1], 1], image.dtype)
    extended = tf.concat([image, extended_channel], axis=2)
    return extended


def unwrap(image: tf.Tensor, replace: int) -> tf.Tensor:
    """Unwraps an image produced by wrap.

    Where there is a 0 in the last channel for every spatial position,
    the rest of the three channels in that spatial dimension are grayed
    (set to 128).  Operations like translate and shear on a wrapped
    Tensor will leave 0s in empty locations.  Some transformations look
    at the intensity of values to do preprocessing, and we want these
    empty pixels to assume the 'average' value, rather than pure black.


    Args:
      image: A 3D Image Tensor with 4 channels.
      replace: A one or three value 1D tensor to fill empty pixels.

    Returns:
      image: A 3D image Tensor with 3 channels.
    """
    image_shape = tf.shape(image)
    # Flatten the spatial dimensions.
    flattened_image = tf.reshape(image, [-1, image_shape[2]])

    # Find all pixels where the last channel is zero.
    alpha_channel = tf.expand_dims(flattened_image[:, 3], axis=-1)

    replace = tf.concat([replace, tf.ones([1], image.dtype)], 0)

    # Where they are zero, fill them in with 'replace'.
    flattened_image = tf.where(tf.equal(alpha_channel, 0), tf.ones_like(flattened_image, dtype=image.dtype) * replace, flattened_image)

    image = tf.reshape(flattened_image, image_shape)
    image = tf.slice(image, [0, 0, 0], [image_shape[0], image_shape[1], 3])
    return image


def _randomly_negate_tensor(tensor):
    """With 50% prob turn the tensor negative."""
    should_flip = tf.cast(tf.floor(tf.random.uniform([]) + 0.5), tf.bool)
    final_tensor = tf.cond(should_flip, lambda: tensor, lambda: -tensor)
    return final_tensor


def _rotate_level_to_arg(level: float):
    level = (level / _MAX_LEVEL) * 30.0
    level = _randomly_negate_tensor(level)
    return (level,)


def _rotate_90_level_to_arg(level: float):
    level = _randomly_negate_tensor(90.0)
    return (level,)


def _shrink_level_to_arg(level: float):
    """Converts level to ratio by which we shrink the image content."""
    if level == 0:
        return (1.0,)  # if level is zero, do not shrink the image
    # Maximum shrinking ratio is 2.9.
    level = 2.0 / (_MAX_LEVEL / level) + 0.9
    return (level,)


def _enhance_level_to_arg(level: float):
    return ((level / _MAX_LEVEL) * 1.8 + 0.1,)  # range [0.1, 1.9]


def _enhance_increasing_level_to_arg(level: float):
    # the 'no change' level is 1.0, moving away from that towards 0. or 2.0 increases the enhancement blend
    # range [0.1, 1.9] if level <= _LEVEL_DENOM
    level = (level / _MAX_LEVEL) * 0.9
    level = 1.0 + _randomly_negate_tensor(level)
    level = tf.reduce_max([level, 0.1])  # keep it >= 0.1
    return (level,)


def _shear_level_to_arg(level: float):
    level = (level / _MAX_LEVEL) * 0.3
    # Flip level to negative with 50% chance.
    level = _randomly_negate_tensor(level)
    return (level,)


def _scale_level_to_arg(level: float):
    level = (level / _MAX_LEVEL) * 0.5
    # Flip level to negative with 50% chance.
    level = _randomly_negate_tensor(level) + 1.0
    return (level,)


def _translate_level_to_arg(level: float, translate_const: float):
    level = (level / _MAX_LEVEL) * float(translate_const)
    # Flip level to negative with 50% chance.
    level = _randomly_negate_tensor(level)
    return (level,)


def _cutout_level_to_arg(level: float, cutout_const: float):
    return (int((level / _MAX_LEVEL) * cutout_const),)


def _posterize_level_to_arg(level: float):
    # As per Tensorflow TPU EfficientNet impl
    # range [0, 4], 'keep 0 up to 4 MSB of original image'
    # intensity/severity of augmentation decreases with level
    return (int((level / _MAX_LEVEL) * 4),)


def _posterize_increasing_level_to_arg(level: float):
    # As per Tensorflow models research and UDA impl
    # range [4, 0], 'keep 4 down to 0 MSB of original image',
    # intensity/severity of augmentation increases with level
    return (4 - int((level / _MAX_LEVEL) * 4),)


def _posterize_original_level_to_arg(level: float):
    # As per original AutoAugment paper description
    # range [4, 8], 'keep 4 up to 8 MSB of image'
    # intensity/severity of augmentation decreases with level
    return (int((level / _MAX_LEVEL) * 4) + 4,)


def _solarize_level_to_arg(level: float):
    # range [0, 256]
    # intensity/severity of augmentation decreases with level
    return (int((level / _MAX_LEVEL) * 256),)


def _solarize_increasing_level_to_arg(level: float):
    # range [0, 256]
    # intensity/severity of augmentation increases with level
    return (256 - int((level / _MAX_LEVEL) * 256),)


def _solarize_add_level_to_arg(level: float):
    # range [0, 110]
    return (int((level / _MAX_LEVEL) * 110),)


def _apply_func_with_prob(func: Any, image: tf.Tensor, args: Any, prob: float):
    """Apply `func` to image w/ `args` as input with probability `prob`."""
    assert isinstance(args, tuple)

    # Apply the function with probability `prob`.
    should_apply_op = tf.cast(tf.floor(tf.random.uniform([], dtype=tf.float32) + prob), tf.bool)
    augmented_image = tf.cond(should_apply_op, lambda: func(image, *args), lambda: image)
    return augmented_image


def select_and_apply_random_policy(policies: Any, image: tf.Tensor):
    """Select a random policy from `policies` and apply it to `image`."""
    policy_to_select = tf.random.uniform([], maxval=len(policies), dtype=tf.int32)
    # Note that using tf.case instead of tf.conds would result in significantly
    # larger graphs and would even break export for some larger policies.
    for (i, policy) in enumerate(policies):
        image = tf.cond(tf.equal(i, policy_to_select), lambda selected_policy=policy: selected_policy(image), lambda: image)
    return image


NAME_TO_FUNC = {
    "AutoContrast": autocontrast,
    "Equalize": equalize,
    "Invert": invert,
    "Rotate": rotate,
    "Rotate90": rotate,
    "Posterize": posterize,
    "PosterizeIncreasing": posterize,
    "PosterizeOriginal": posterize,
    "Solarize": solarize,
    "SolarizeIncreasing": solarize,
    "SolarizeAdd": solarize_add,
    "Color": color,
    "ColorIncreasing": color,
    "Contrast": contrast,
    "ContrastIncreasing": contrast,
    "Brightness": brightness,
    "BrightnessIncreasing": brightness,
    "Sharpness": sharpness,
    "SharpnessIncreasing": sharpness,
    "ShearX": shear_x,
    "ShearY": shear_y,
    "ScaleX": scale_x,
    "ScaleY": scale_y,
    "Scale": scale,
    "TranslateX": translate_x,
    "TranslateY": translate_y,
    "TranslateXRel": translate_x_relative,
    "TranslateYRel": translate_y_relative,
    "Cutout": cutout,
}

# Functions that have a 'replace' parameter
REPLACE_FUNCS = frozenset(
    {
        "Rotate",
        "Rotate90",
        "TranslateX",
        "TranslateXRel",
        "ShearX",
        "ShearY",
        "ScaleX",
        "ScaleY",
        "Scale",
        "TranslateY",
        "TranslateYRel",
        "Cutout",
    }
)

LEVEL_TO_ARG = {
    "AutoContrast": lambda level: (),
    "Equalize": lambda level: (),
    "Invert": lambda level: (),
    "Rotate": _rotate_level_to_arg,
    "Rotate90": _rotate_90_level_to_arg,
    # There are several variations of the posterize level scaling in various Tensorflow/Google repositories/papers
    "Posterize": _posterize_level_to_arg,
    "PosterizeIncreasing": _posterize_increasing_level_to_arg,
    "PosterizeOriginal": _posterize_original_level_to_arg,
    "Solarize": _solarize_level_to_arg,
    "SolarizeIncreasing": _solarize_increasing_level_to_arg,
    "SolarizeAdd": _solarize_add_level_to_arg,
    "Color": _enhance_level_to_arg,
    "ColorIncreasing": _enhance_increasing_level_to_arg,
    "Contrast": _enhance_level_to_arg,
    "ContrastIncreasing": _enhance_increasing_level_to_arg,
    "Brightness": _enhance_level_to_arg,
    "BrightnessIncreasing": _enhance_increasing_level_to_arg,
    "Sharpness": _enhance_level_to_arg,
    "SharpnessIncreasing": _enhance_increasing_level_to_arg,
    "ShearX": _shear_level_to_arg,
    "ShearY": _shear_level_to_arg,
    "ScaleX": _scale_level_to_arg,
    "ScaleY": _scale_level_to_arg,
    "Scale": _scale_level_to_arg,
    "TranslateX": _translate_level_to_arg,
    "TranslateY": _translate_level_to_arg,
    "TranslateXRel": _translate_level_to_arg,
    "TranslateYRel": _translate_level_to_arg,
    "Cutout": _cutout_level_to_arg,
}


def _parse_policy_info(name: Text, prob: float, level: float, replace_value: List[int], cutout_const: float, translate_const: float) -> Tuple[Any, float, Any]:
    """Return the function that corresponds to `name` and update `level` param."""
    func = NAME_TO_FUNC[name]
    level_func = LEVEL_TO_ARG[name]
    if name == "Cutout":
        args = level_func(level, cutout_const)
    elif name in ["TranslateX", "TranslateY", "TranslateXRel", "TranslateYRel"]:
        args = level_func(level, translate_const)
    else:
        args = level_func(level)

    if name in REPLACE_FUNCS:
        # Add in replace arg if it is required for the function that is called.
        args = tuple(list(args) + [replace_value])

    return func, prob, args


class ImageAugment(object):
    """Image augmentation class for applying image distortions."""

    def distort(self, image: tf.Tensor) -> tf.Tensor:
        """Given an image tensor, returns a distorted image with the same shape.

        Args:
          image: `Tensor` of shape [height, width, 3] representing an image.

        Returns:
          The augmented version of `image`.
        """
        return self.__call__(image)


class AutoAugment(ImageAugment):
    """Applies the AutoAugment policy to images.

    AutoAugment is from the paper: https://arxiv.org/abs/1805.09501.
    """

    def __init__(
        self,
        augmentation_name: Text = "v0",
        policies: Optional[Dict[Text, Any]] = None,
        cutout_const: float = 100,
        translate_const: float = 250,
    ):
        """Applies the AutoAugment policy to images.

        Args:
          augmentation_name: The name of the AutoAugment policy to use. The
            available options are `v0` and `test`. `v0` is the policy used for all
            of the results in the paper and was found to achieve the best results on
            the COCO dataset. `v1`, `v2` and `v3` are additional good policies found
            on the COCO dataset that have slight variation in what operations were
            used during the search procedure along with how many operations are
            applied in parallel to a single image (2 vs 3).
          policies: list of lists of tuples in the form `(func, prob, level)`,
            `func` is a string name of the augmentation function, `prob` is the
            probability of applying the `func` operation, `level` is the input
            argument for `func`.
          cutout_const: multiplier for applying cutout.
          translate_const: multiplier for applying translation.
        """
        super(AutoAugment, self).__init__()

        if policies is None:
            self.available_policies = {
                "v0": self.policy_v0(),
                "test": self.policy_test(),
                "simple": self.policy_simple(),
            }

        if augmentation_name not in self.available_policies:
            raise ValueError("Invalid augmentation_name: {}".format(augmentation_name))

        self.augmentation_name = augmentation_name
        self.policies = self.available_policies[augmentation_name]
        self.cutout_const = float(cutout_const)
        self.translate_const = float(translate_const)

    def __call__(self, image: tf.Tensor) -> tf.Tensor:
        """Applies the AutoAugment policy to `image`.

        AutoAugment is from the paper: https://arxiv.org/abs/1805.09501.

        Args:
          image: `Tensor` of shape [height, width, 3] representing an image.

        Returns:
          A version of image that now has data augmentation applied to it based on
          the `policies` pass into the function.
        """
        input_image_type = image.dtype

        if input_image_type != tf.uint8:
            image = tf.clip_by_value(image, 0.0, 255.0)
            image = tf.cast(image, dtype=tf.uint8)

        replace_value = [128] * 3

        # func is the string name of the augmentation function, prob is the
        # probability of applying the operation and level is the parameter
        # associated with the tf op.

        # tf_policies are functions that take in an image and return an augmented
        # image.
        tf_policies = []
        for policy in self.policies:
            tf_policy = []
            # Link string name to the correct python function and make sure the
            # correct argument is passed into that function.
            for policy_info in policy:
                policy_info = list(policy_info) + [replace_value, self.cutout_const, self.translate_const]
                tf_policy.append(_parse_policy_info(*policy_info))
            # Now build the tf policy that will apply the augmentation procedue
            # on image.
            def make_final_policy(tf_policy_):
                def final_policy(image_):
                    for func, prob, args in tf_policy_:
                        image_ = _apply_func_with_prob(func, image_, args, prob)
                    return image_

                return final_policy

            tf_policies.append(make_final_policy(tf_policy))

        image = select_and_apply_random_policy(tf_policies, image)
        image = tf.cast(image, dtype=input_image_type)
        return image

    @staticmethod
    def policy_v0():
        """Autoaugment policy that was used in AutoAugment Paper.

        Each tuple is an augmentation operation of the form
        (operation, probability, magnitude). Each element in policy is a
        sub-policy that will be applied sequentially on the image.

        Returns:
          the policy.
        """

        # TODO(dankondratyuk): tensorflow_addons defines custom ops, which
        # for some reason are not included when building/linking
        # This results in the error, "Op type not registered
        # 'Addons>ImageProjectiveTransformV2' in binary" when running on borg TPUs
        policy = [
            [("Color", 0.4, 9), ("Equalize", 0.6, 3)],
            [("Solarize", 0.8, 3), ("Equalize", 0.4, 7)],
            [("Solarize", 0.4, 2), ("Solarize", 0.6, 2)],
            [("Color", 0.2, 0), ("Equalize", 0.8, 8)],
            [("Equalize", 0.4, 8), ("SolarizeAdd", 0.8, 3)],
            [("Color", 0.6, 1), ("Equalize", 1.0, 2)],
            [("Color", 0.4, 7), ("Equalize", 0.6, 0)],
            [("Posterize", 0.4, 6), ("AutoContrast", 0.4, 7)],
            [("Solarize", 0.6, 8), ("Color", 0.6, 9)],
            [("Equalize", 0.8, 4), ("Equalize", 0.0, 8)],
            [("Equalize", 1.0, 4), ("AutoContrast", 0.6, 2)],
            [("Posterize", 0.8, 2), ("Solarize", 0.6, 10)],
            [("Solarize", 0.6, 8), ("Equalize", 0.6, 1)],
            [("Equalize", 0.8, 1), ("ShearY", 0.8, 4)],
            [("Color", 0.4, 1), ("Rotate", 0.6, 8)],
            [("ShearX", 0.2, 9), ("Rotate", 0.6, 8)],
            [("Invert", 0.4, 9), ("Rotate", 0.6, 0)],
            [("Equalize", 1.0, 9), ("ShearY", 0.6, 3)],
            [("Solarize", 0.2, 4), ("Rotate", 0.8, 9)],
            [("Rotate", 1.0, 7), ("TranslateY", 0.8, 9)],
            [("ShearX", 0.0, 0), ("Solarize", 0.8, 4)],
            [("ShearY", 0.8, 0), ("Color", 0.6, 4)],
            [("Color", 1.0, 0), ("Rotate", 0.6, 2)],
            [("ShearY", 0.4, 7), ("SolarizeAdd", 0.6, 7)],
            [("Color", 0.8, 6), ("Rotate", 0.4, 5)],
        ]
        return policy

    @staticmethod
    def policy_simple():
        """Same as `policy_v0`, except with custom ops removed."""

        policy = [
            [("Color", 0.4, 9), ("Equalize", 0.6, 3)],
            [("Solarize", 0.8, 3), ("Equalize", 0.4, 7)],
            [("Solarize", 0.4, 2), ("Solarize", 0.6, 2)],
            [("Color", 0.2, 0), ("Equalize", 0.8, 8)],
            [("Equalize", 0.4, 8), ("SolarizeAdd", 0.8, 3)],
            [("Color", 0.6, 1), ("Equalize", 1.0, 2)],
            [("Color", 0.4, 7), ("Equalize", 0.6, 0)],
            [("Posterize", 0.4, 6), ("AutoContrast", 0.4, 7)],
            [("Solarize", 0.6, 8), ("Color", 0.6, 9)],
            [("Equalize", 0.8, 4), ("Equalize", 0.0, 8)],
            [("Equalize", 1.0, 4), ("AutoContrast", 0.6, 2)],
            [("Posterize", 0.8, 2), ("Solarize", 0.6, 10)],
            [("Solarize", 0.6, 8), ("Equalize", 0.6, 1)],
        ]
        return policy

    @staticmethod
    def policy_positional():
        policy = [
            [("Equalize", 0.8, 1), ("ShearY", 0.8, 4)],
            [("Color", 0.4, 1), ("Rotate", 0.6, 8)],
            [("ShearX", 0.2, 9), ("Rotate", 0.6, 8)],
            [("Invert", 0.4, 9), ("Rotate", 0.6, 0)],
            [("Equalize", 1.0, 9), ("ShearY", 0.6, 3)],
            [("Solarize", 0.2, 4), ("Rotate", 0.8, 9)],
            [("Rotate", 1.0, 7), ("TranslateY", 0.8, 9)],
            [("ShearX", 0.0, 0), ("Solarize", 0.8, 4)],
            [("ShearY", 0.8, 0), ("Color", 0.6, 4)],
            [("Color", 1.0, 0), ("Rotate", 0.6, 2)],
            [("ShearY", 0.4, 7), ("SolarizeAdd", 0.6, 7)],
            [("Color", 0.8, 6), ("Rotate", 0.4, 5)],
        ]
        return policy

    @staticmethod
    def policy_test():
        """Autoaugment test policy for debugging."""
        policy = [
            [("TranslateX", 1.0, 4), ("Equalize", 1.0, 10)],
        ]
        return policy


class RandAugment(ImageAugment):
    """Applies the RandAugment policy to images.

    RandAugment is from the paper https://arxiv.org/abs/1909.13719,
    """

    def __init__(
        self,
        num_layers: int = 2,
        magnitude: float = 10.0,
        magnitude_max: float = _MAX_LEVEL,
        magnitude_std: float = 0.5,
        cutout_const: float = 40.0,
        translate_const: float = 0.45,  # (0, 1) for relative, > 1 for absolute.
        use_cutout: bool = False,
        use_relative_translate: bool = True,
        use_color_increasing: bool = True,
        use_positional_related_ops: bool = True,  # Set False to exlude [shear, rotate, translate]
        apply_probability: float = 0.5,
        image_mean: Union[list, tuple] = [124, 117, 104],
    ):
        """Applies the RandAugment policy to images.

        Args:
          num_layers: Integer, the number of augmentation transformations to apply
            sequentially to an image. Represented as (N) in the paper. Usually best
            values will be in the range [1, 3].
          magnitude: Integer, shared magnitude across all augmentation operations.
            Represented as (M) in the paper. Usually best values are in the range
            [5, 10].
          cutout_const: multiplier for applying cutout.
          translate_const: multiplier for applying translation.
        """
        super(RandAugment, self).__init__()

        self.num_layers, self.apply_probability, self.image_mean = num_layers, apply_probability, image_mean
        self.magnitude, self.magnitude_max, self.magnitude_std = float(magnitude), float(magnitude_max), float(magnitude_std)
        self.cutout_const, self.translate_const = float(cutout_const), float(translate_const)
        self.basic_ops = [
            "AutoContrast",
            "Equalize",
            "Invert",
            "SolarizeAdd",
        ]
        self.color_ops = ["Posterize", "Solarize", "Color", "Contrast", "Brightness", "Sharpness"]
        self.color_increasing_ops = [
            "PosterizeIncreasing",
            "SolarizeIncreasing",
            "ColorIncreasing",
            "ContrastIncreasing",
            "BrightnessIncreasing",
            "SharpnessIncreasing",
        ]
        self.positional_related_ops = ["Rotate", "ShearX", "ShearY"]
        self.positional_related_ops += ["TranslateXRel", "TranslateYRel"] if use_relative_translate else ["TranslateX", "TranslateY"]

        self.available_ops = self.basic_ops
        self.available_ops += self.color_increasing_ops if use_color_increasing else self.color_ops
        self.available_ops += ["Cutout"] if use_cutout else []
        if use_positional_related_ops:
            self.available_ops += self.positional_related_ops

    def __magnitude_with_noise__(self):
        if self.magnitude_std > 0:
            magnitude = tf.random.normal((), mean=self.magnitude, stddev=self.magnitude_std)
        else:
            magnitude = self.magnitude
        return tf.clip_by_value(magnitude, 0, self.magnitude_max)

    def apply_policy(self, policy, image):
        # print(f">>>> {policy = }")
        magnitude = self.__magnitude_with_noise__()
        # policy = random.choice(self.available_ops)
        func, _, args = _parse_policy_info(policy, 0.0, magnitude, self.image_mean, self.cutout_const, self.translate_const)
        return func(image, *args)

    def select_and_apply_random_policy(self, image):
        """Select a random policy from `policies` and apply it to `image`."""
        policy_to_select = tf.random.uniform([], maxval=len(self.available_ops), dtype=tf.int32)
        branch_fns = [(id, lambda selected=ii: self.apply_policy(selected, image)) for id, ii in enumerate(self.available_ops)]
        image = tf.switch_case(branch_index=policy_to_select, branch_fns=branch_fns, default=lambda: tf.identity(image))
        # Note that using tf.case instead of tf.conds would result in significantly
        # larger graphs and would even break export for some larger policies.
        # for (i, policy) in enumerate(self.available_ops):
        #     image = tf.cond(tf.equal(i, policy_to_select), lambda policy=policy: self.apply_policy(policy, image), lambda: image)
        return image

    def __call__(self, image: tf.Tensor) -> tf.Tensor:
        """Applies the RandAugment policy to `image`.

        Args:
          image: `Tensor` of shape [height, width, 3] representing an image.

        Returns:
          The augmented version of `image`.
        """
        input_image_type = image.dtype

        if input_image_type != tf.uint8:
            image = tf.clip_by_value(image, 0.0, 255.0)
            image = tf.cast(image, dtype=tf.uint8)

        for _ in range(self.num_layers):
            should_apply_op = tf.cast(tf.floor(tf.random.uniform([], dtype=tf.float32) + self.apply_probability), tf.bool)
            image = tf.cond(should_apply_op, lambda: self.select_and_apply_random_policy(image), lambda: image)
        image = tf.cast(image, dtype=input_image_type)
        return image


class PositionalRandAugment(RandAugment):
    """Applies the RandAugment positional related policy to images. Including [shear, rotate, translate], Also returns affine transform matrix"""

    def __init__(
        self,
        num_layers: int = 2,
        magnitude: float = 10.0,
        magnitude_max: float = _MAX_LEVEL,
        magnitude_std: float = 0.5,
        translate_const: float = 0.45,  # (0, 1) for relative, > 1 for absolute.
        use_relative_translate: bool = True,
        apply_probability: float = 0.5,
        image_mean: Union[list, tuple] = [124, 117, 104],
        positional_augment_methods: str = "rts",
        **kwargs,  # Not using, just in case any additional params.
    ):
        self.num_layers, self.apply_probability, self.image_mean = num_layers, apply_probability, image_mean
        self.magnitude, self.magnitude_max, self.magnitude_std = float(magnitude), float(magnitude_max), float(magnitude_std)
        self.translate_const = float(translate_const)

        positional_augment_methods = "" if positional_augment_methods is None else positional_augment_methods
        self.available_ops = ["ScaleX", "ScaleY"] if "x" in positional_augment_methods else ["Scale"]
        if "r" in positional_augment_methods:
            self.available_ops += ["Rotate"]
        if "R" in positional_augment_methods:
            self.available_ops += ["Rotate90"]
        if "s" in positional_augment_methods:
            self.available_ops += ["ShearX", "ShearY"]
        if "t" in positional_augment_methods:
            self.available_ops += ["TranslateXRel", "TranslateYRel"] if use_relative_translate else ["TranslateX", "TranslateY"]
        self.DEFAULT_AFFINE = tf.constant([[1.0, 0, 0, 0, 1, 0, 0, 0]])

    def apply_policy(self, policy, image):
        # print(f">>>> {policy = }")
        magnitude = self.__magnitude_with_noise__()
        func, _, args = _parse_policy_info(policy, 0.0, magnitude, self.image_mean, 0, self.translate_const)
        return func(image, *args, return_affine_matrix=True)

    def select_and_apply_random_policy(self, image):
        """Select a random policy from `policies` and apply it to `image`."""
        policy_to_select = tf.random.uniform([], maxval=len(self.available_ops), dtype=tf.int32)
        branch_fns = [(id, lambda selected=ii: self.apply_policy(selected, image)) for id, ii in enumerate(self.available_ops)]
        return tf.switch_case(branch_index=policy_to_select, branch_fns=branch_fns, default=lambda: self.DEFAULT_AFFINE)

    def __call__(self, image: tf.Tensor) -> tf.Tensor:
        input_image_type = image.dtype

        if input_image_type != tf.uint8:
            image = tf.clip_by_value(image, 0.0, 255.0)
            image = tf.cast(image, dtype=tf.uint8)

        result_affine_matrix = self.DEFAULT_AFFINE
        for _ in range(self.num_layers):
            should_apply_op = tf.cast(tf.floor(tf.random.uniform([], dtype=tf.float32) + self.apply_probability), tf.bool)
            affine_matrix = tf.cond(should_apply_op, lambda: self.select_and_apply_random_policy(image), lambda: self.DEFAULT_AFFINE)
            result_affine_matrix = marge_two_transforms(result_affine_matrix, affine_matrix)

        image = transform(image=wrap(image), transforms=result_affine_matrix)
        image = unwrap(image, self.image_mean)
        image = tf.cast(image, dtype=input_image_type)
        return image, result_affine_matrix



import tensorflow as tf
from tensorflow import keras
from keras_cv_attention_models.imagenet.data import init_mean_std_by_rescale_mode, tf_imread, random_crop_and_resize_image
from keras_cv_attention_models.coco import anchors_func

COCO_LABELS = """person, bicycle, car, motorcycle, airplane, bus, train, truck, boat, traffic light, fire hydrant, stop sign,
    parking meter, bench, bird, cat, dog, horse, sheep, cow, elephant, bear, zebra, giraffe, backpack, umbrella, handbag, tie,
    suitcase, frisbee, skis, snowboard, sports ball, kite, baseball bat, baseball glove, skateboard, surfboard, tennis racket,
    bottle, wine glass, cup, fork, knife, spoon, bowl, banana, apple, sandwich, orange, broccoli, carrot, hot dog, pizza, donut,
    cake, chair, couch, potted plant, bed, dining table, toilet, tv, laptop, mouse, remote, keyboard, cell phone, microwave, oven,
    toaster, sink, refrigerator, book, clock, vase, scissors, teddy bear, hair drier, toothbrush"""
COCO_80_LABEL_DICT = {id: ii.strip() for id, ii in enumerate(COCO_LABELS.split(","))}
INVALID_ID_90 = [11, 25, 28, 29, 44, 65, 67, 68, 70, 82]
COCO_90_LABEL_DICT = {id: ii for id, ii in zip(set(range(90)) - set(INVALID_ID_90), COCO_80_LABEL_DICT.values())}
COCO_90_LABEL_DICT.update({ii: "Unknown" for ii in INVALID_ID_90})
COCO_80_to_90_LABEL_DICT = {id_80: id_90 for id_80, id_90 in enumerate(set(range(90)) - set(INVALID_ID_90))}


""" Bboxes augment """


def rerange_scale_offset_to_01(source_height, source_width, target_height, target_width, scale_hh, scale_ww, offset_hh=0, offset_ww=0):
    # Input: image size firstly rescale with (scale_hh, scale_ww), then crop as [offset_hh: offset_hh + target_height, offset_ww: offset_ww + target_width]
    # Output: coords or bbox value in (0, 1) -> * scale - offset
    source_height, source_width = tf.cast(source_height, "float32"), tf.cast(source_width, "float32")
    target_height, target_width = tf.cast(target_height, "float32"), tf.cast(target_width, "float32")
    scale_hh_01, scale_ww_01 = source_height * scale_hh / target_height, source_width * scale_ww / target_width
    crop_hh_01, crop_ww_01 = tf.cast(offset_hh, "float32") / target_height, tf.cast(offset_ww, "float32") / target_width
    return scale_hh_01, scale_ww_01, crop_hh_01, crop_ww_01


def inverse_affine_matrix_single_6(affine):
    scale_inverse = affine[0] * affine[4] - affine[3] * affine[1]
    affine_aa = [affine[4], -affine[1], affine[5] * affine[1] - affine[2] * affine[4]]
    affine_bb = [-affine[3], affine[0], affine[2] * affine[3] - affine[5] * affine[0]]
    return tf.convert_to_tensor(affine_aa + affine_bb, dtype=affine.dtype) / scale_inverse


def bboxes_apply_affine(bboxes, affine, input_shape, inverse=True):
    height, width = tf.cast(input_shape[0], bboxes.dtype), tf.cast(input_shape[1], bboxes.dtype)
    top, left, bottom, right = bboxes[:, 0] * height, bboxes[:, 1] * width, bboxes[:, 2] * height, bboxes[:, 3] * width
    affine = tf.squeeze(affine)[:6]
    if inverse:
        affine = inverse_affine_matrix_single_6(affine)
    affine = tf.reshape(affine, [2, 3])

    # 4 corners points: [left, top, right, bottom, left, bottom, right, top]
    corners = tf.gather(bboxes * [height, width, height, width], [1, 0, 3, 2, 1, 2, 3, 0], axis=-1)
    # -> [[left, top], [right, bottom], [left, bottom], [right, top]]
    corners = tf.reshape(corners, [-1, 4, 2])
    # pad 1
    corners = tf.concat([corners, tf.ones_like(corners[:, :, :1])], axis=-1)
    # apply affine transform
    corners_transformed = tf.matmul(corners, affine, transpose_b=True)

    new_left = tf.minimum(corners_transformed[:, 0, 0], corners_transformed[:, 2, 0]) / width
    new_top = tf.minimum(corners_transformed[:, 0, 1], corners_transformed[:, 3, 1]) / height
    new_right = tf.maximum(corners_transformed[:, 1, 0], corners_transformed[:, 3, 0]) / width
    new_bottom = tf.maximum(corners_transformed[:, 1, 1], corners_transformed[:, 2, 1]) / height
    return tf.stack([new_top, new_left, new_bottom, new_right], axis=-1)


def refine_bboxes_labels_single(bboxes, labels):
    bboxes = tf.clip_by_value(bboxes, 0, 1)
    picking_indices = tf.where(tf.not_equal((bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1]), 0))
    bboxes, labels = tf.gather_nd(bboxes, picking_indices), tf.gather_nd(labels, picking_indices)
    return bboxes, labels


""" Image Augment """


def random_largest_crop_and_resize_images(images, target_shape, method="bilinear", antialias=False):
    if tf.rank(images) == 3:
        height, width = tf.cast(tf.shape(images)[0], "float32"), tf.cast(tf.shape(images)[1], "float32")
    else:
        height, width = tf.cast(tf.shape(images)[1], "float32"), tf.cast(tf.shape(images)[2], "float32")
    target_height, target_width = tf.cast(target_shape[0], "float32"), tf.cast(target_shape[1], "float32")
    scale = tf.maximum(target_height / height, target_width / width)
    scaled_hh, scaled_ww = tf.cast(tf.math.ceil(height * scale), "int32"), tf.cast(tf.math.ceil(width * scale), "int32")
    images = tf.image.resize(images, [scaled_hh, scaled_ww], method=method, antialias=antialias)

    # print(target_shape)
    crop_hh = tf.cond(scaled_hh > target_shape[0], lambda: tf.random.uniform((), 0, scaled_hh - target_shape[0], dtype="int32"), lambda: 0)
    crop_ww = tf.cond(scaled_ww > target_shape[1], lambda: tf.random.uniform((), 0, scaled_ww - target_shape[1], dtype="int32"), lambda: 0)
    # crop_hh = tf.random.uniform((), 0, tf.maximum(scaled_hh - target_shape[0], 1), dtype='int32')
    # crop_ww = tf.random.uniform((), 0, tf.maximum(scaled_ww - target_shape[1], 1), dtype='int32')
    if tf.rank(images) == 3:
        images = images[crop_hh : crop_hh + target_shape[0], crop_ww : crop_ww + target_shape[1]]
    else:
        images = images[:, crop_hh : crop_hh + target_shape[0], crop_ww : crop_ww + target_shape[1]]
    return images, scale, crop_hh, crop_ww


def get_image_aspect_aware_random_scale_crop(source_shape, target_shape, scale_min=0.1, scale_max=2.0):
    """https://github.com/google/automl/tree/master/efficientdet/dataloader.py#L67"""
    random_scale_factor = tf.random.uniform((), scale_min, scale_max)
    scaled_y, scaled_x = random_scale_factor * target_shape[0], random_scale_factor * target_shape[1]
    height, width = tf.cast(source_shape[0], tf.float32), tf.cast(source_shape[1], tf.float32)
    random_image_scale = tf.minimum(scaled_y / height, scaled_x / width)

    # Select non-zero random offset (x, y) if scaled image is larger than self._output_size.
    height, width = tf.cast(source_shape[0], tf.float32), tf.cast(source_shape[1], tf.float32)
    scaled_height, scaled_width = height * random_image_scale, width * random_image_scale
    offset_y, offset_x = tf.maximum(0.0, scaled_height - target_shape[0]), tf.maximum(0.0, scaled_width - target_shape[1])
    random_offset_y, random_offset_x = offset_y * tf.random.uniform([], 0, 1), offset_x * tf.random.uniform([], 0, 1)
    random_offset_y, random_offset_x = tf.cast(random_offset_y, tf.int32), tf.cast(random_offset_x, tf.int32)
    return random_image_scale, random_offset_y, random_offset_x


def aspect_aware_resize_and_crop_image(image, target_shape, scale=-1, crop_y=0, crop_x=0, letterbox_pad=-1, method="bilinear", antialias=False):
    letterbox_target_shape = (target_shape[0] - letterbox_pad, target_shape[1] - letterbox_pad) if letterbox_pad > 0 else target_shape
    height, width = tf.cast(tf.shape(image)[0], "float32"), tf.cast(tf.shape(image)[1], "float32")
    if scale == -1:
        scale = tf.minimum(letterbox_target_shape[0] / height, letterbox_target_shape[1] / width)
    scaled_hh, scaled_ww = int(height * scale), int(width * scale)
    image = tf.image.resize(image, [scaled_hh, scaled_ww], method=method, antialias=antialias)
    image = image[crop_y : crop_y + letterbox_target_shape[0], crop_x : crop_x + letterbox_target_shape[1]]
    cropped_shape = tf.shape(image)

    pad_top, pad_left = ((target_shape[0] - cropped_shape[0]) // 2, (target_shape[1] - cropped_shape[1]) // 2) if letterbox_pad >= 0 else (0, 0)
    image = tf.image.pad_to_bounding_box(image, pad_top, pad_left, target_shape[0], target_shape[1])
    return image, scale, pad_top, pad_left


def random_flip_left_right_with_bboxes(image, bboxes, probability=0.5):
    # For box, left = 1 - right, right = 1 - left
    return tf.cond(
        tf.random.uniform(()) < probability,
        lambda: (tf.image.flip_left_right(image), tf.gather(bboxes, [0, 3, 2, 1], axis=1) * [1, -1, 1, -1] + [0, 1, 0, 1]),
        lambda: (image, bboxes),
    )


def random_hsv(image, hue_delta=0.015, saturation_delta=0.7, brightness_delta=0.4, contrast_delta=0, show_sample=False):
    # augment_hsv https://github.com/WongKinYiu/yolor/blob/main/utils/datasets.py#L941
    if show_sample:
        import matplotlib.pyplot as plt

        fig = plt.figure()
        aa = tf.concat([tf.image.adjust_hue(image, -hue_delta), tf.image.adjust_hue(image, hue_delta)], axis=1)
        bb = tf.concat([tf.image.adjust_saturation(image, 1 - saturation_delta), tf.image.adjust_saturation(image, 1 + saturation_delta)], axis=1)
        cc = tf.concat([tf.image.adjust_brightness(image, -brightness_delta), tf.image.adjust_brightness(image, brightness_delta)], axis=1)
        dd = tf.concat([tf.image.adjust_contrast(image, 1 - contrast_delta), tf.image.adjust_contrast(image, 1 + contrast_delta)], axis=1)
        plt.imshow(tf.concat([aa, bb, cc, dd], axis=0))
        plt.axis("off")
        plt.tight_layout()

    image = tf.image.random_hue(image, hue_delta)
    image = tf.image.random_saturation(image, 1 - saturation_delta, 1 + saturation_delta)
    image = tf.image.random_brightness(image, brightness_delta)
    if contrast_delta > 0:
        image = tf.image.random_contrast(image, 1 - contrast_delta, 1 + contrast_delta)
    return image


""" Mosaic mix """


def mosaic_mix_batch(images, bboxes, labels, split_center_min=0.25, split_center_max=0.75):
    batch_size = tf.shape(images)[0]
    _, hh, ww, _ = images.shape
    split_hh = tf.cast(tf.random.uniform((), split_center_min * hh, split_center_max * hh), "int32")
    split_ww = tf.cast(tf.random.uniform((), split_center_min * ww, split_center_max * ww), "int32")
    # print(split_hh, split_ww)

    # top_left, top_right, bottom_left, bottom_right
    # sub_hh_wws = [[split_hh, split_ww], [split_hh, ww - split_ww], [hh - split_hh, split_ww], [hh - split_hh, ww - split_ww]]
    starts = [[0, 0], [0, split_ww], [split_hh, 0], [split_hh, split_ww]]
    ends = [[split_hh, split_ww], [split_hh, ww], [hh, split_ww], [hh, ww]]
    mixed_images, mixed_bboxes, mixed_labels = [], [], []
    for (top, left), (bottom, right) in zip(starts, ends):
        sub_hh, sub_ww = bottom - top, right - left
        pick_indices = tf.random.shuffle(tf.range(batch_size))
        cur_images = tf.gather(images, pick_indices)
        cur_images, scale, crop_hh, crop_ww = random_largest_crop_and_resize_images(cur_images, [sub_hh, sub_ww])
        mixed_images.append(cur_images)
        # print(f"{cur_images.shape = }, {scale = }")

        cur_bboxes, cur_labels = tf.gather(bboxes, pick_indices), tf.gather(labels, pick_indices)
        target_height, target_width = 1, 1  # Don't scale back to [0, 1]
        scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(hh, ww, target_height, target_width, scale, scale, crop_hh, crop_ww)
        cur_bboxes = cur_bboxes * [scale_hh, scale_ww, scale_hh, scale_ww] - [crop_hh, crop_ww, crop_hh, crop_ww]

        """ Re-fine batch bboxes """
        sub_hh, sub_ww = tf.cast(sub_hh, cur_bboxes.dtype), tf.cast(sub_ww, cur_bboxes.dtype)
        cur_bboxes = tf.clip_by_value(cur_bboxes, 0, [sub_hh, sub_ww, sub_hh, sub_ww])
        mask_cond = tf.not_equal((cur_bboxes[:, :, 2] - cur_bboxes[:, :, 0]) * (cur_bboxes[:, :, 3] - cur_bboxes[:, :, 1]), 0)
        cur_bboxes += [top, left, top, left]
        cur_bboxes = cur_bboxes * tf.expand_dims(tf.cast(mask_cond, cur_bboxes.dtype), -1)
        cur_labels = tf.where(mask_cond, cur_labels, tf.zeros_like(cur_labels))

        mixed_bboxes.append(cur_bboxes)
        mixed_labels.append(cur_labels)

    top_images = tf.concat([mixed_images[0], mixed_images[1]], axis=2)
    bottom_images = tf.concat([mixed_images[2], mixed_images[3]], axis=2)
    mixed_images = tf.concat([top_images, bottom_images], axis=1)
    mixed_images.set_shape([None, hh, ww, 3])
    mixed_bboxes = tf.concat(mixed_bboxes, axis=1) / [hh, ww, hh, ww]
    mixed_labels = tf.concat(mixed_labels, axis=1)
    # print(f"{top_images.shape = }, {bottom_images.shape = }, {mix.shape = }, {mixed_images.shape = }, {mixed_labels.shape = }")
    return mixed_images, (mixed_bboxes, mixed_labels)


""" Dataset """


class RandomProcessImageWithBboxes:
    def __init__(
        self,
        target_shape=(300, 300),
        max_labels_per_image=100,
        random_crop_mode=0,  # 0 for eval mode, (0, 1) for random crop, 1 for random largest crop, > 1 for random scale
        resize_method="bilinear",
        resize_antialias=False,
        color_augment_method="random_hsv",
        magnitude=0,
        num_layers=2,
        use_color_increasing=True,
        **randaug_kwargs,
    ):
        self.max_labels_per_image = max_labels_per_image
        self.target_shape = target_shape if len(target_shape) == 2 else target_shape[:2]
        self.resize_method, self.resize_antialias, self.random_crop_mode, self.magnitude = resize_method, resize_antialias, random_crop_mode, magnitude
        if magnitude > 0:
            from keras_cv_attention_models.imagenet import augment

            if color_augment_method is None:
                self.randaug_wo_pos = lambda image: image
            elif isinstance(color_augment_method, str) and color_augment_method.lower() == "autoaug":
                self.randaug_wo_pos = augment.AutoAugment(augmentation_name="simple")
            elif isinstance(color_augment_method, str) and color_augment_method.lower() == "randaug":
                # TODO: Need to pick color related methods, "Invert" / "Posterize" may not working well here.
                print(">>>> RandAugment: magnitude = %d" % magnitude)
                self.randaug_wo_pos = augment.RandAugment(
                    num_layers=num_layers,
                    magnitude=magnitude,
                    use_cutout=False,
                    use_color_increasing=use_color_increasing,
                    use_positional_related_ops=False,  # Set False to exlude [shear, rotate, translate]
                    **randaug_kwargs,
                )
            elif isinstance(color_augment_method, str):
                self.randaug_wo_pos = random_hsv
            else:
                print(">>>> Using user defined color_augment_method: {}".format(getattr(color_augment_method, "__name__", color_augment_method)))
                self.randaug_wo_pos = color_augment_method

            # RandAugment positional related ops. Including [shear, rotate, translate], Also returns affine transform matrix
            # self.pos_randaug = augment.PositionalRandAugment(num_layers=num_layers, magnitude=magnitude, **randaug_kwargs)

    def __random_crop_and_resize__(self, image, bbox):
        height, width = tf.shape(image)[0], tf.shape(image)[1]

        if self.random_crop_mode > 1:
            scale, crop_hh, crop_ww = get_image_aspect_aware_random_scale_crop((height, width), self.target_shape, scale_max=self.random_crop_mode)
            image, scale, _, _ = aspect_aware_resize_and_crop_image(
                image, self.target_shape, scale, crop_hh, crop_ww, method=self.resize_method, antialias=self.resize_antialias
            )
            scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(
                height, width, self.target_shape[0], self.target_shape[1], scale, scale, crop_hh, crop_ww
            )
        elif self.random_crop_mode == 1:
            image, scale, crop_hh, crop_ww = random_largest_crop_and_resize_images(image, self.target_shape, self.resize_method, self.resize_antialias)
            scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(
                height, width, self.target_shape[0], self.target_shape[1], scale, scale, crop_hh, crop_ww
            )
        elif self.random_crop_mode > 0 and self.random_crop_mode < 1:
            image, scale_hh, scale_ww, crop_hh, crop_ww = random_crop_and_resize_image(
                image, self.target_shape, scale=(self.random_crop_mode, 1.0), method=self.resize_method, antialias=self.resize_antialias
            )
        else:
            image, scale, _, _ = aspect_aware_resize_and_crop_image(image, self.target_shape, method=self.resize_method, antialias=self.resize_antialias)
            scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(height, width, self.target_shape[0], self.target_shape[1], scale, scale)

        bbox = bbox * [scale_hh, scale_ww, scale_hh, scale_ww] - [crop_hh, crop_ww, crop_hh, crop_ww]
        return image, bbox

    def __call__(self, datapoint):
        image = datapoint["image"]
        objects = datapoint["objects"]
        bbox, label, is_crowd = tf.cast(objects["bbox"], tf.float32), objects["label"], objects.get("is_crowd", None)
        if len(image.shape) < 2:
            image = tf_imread(image)
        if is_crowd is not None:
            is_not_crowd = is_crowd == False
            bbox, label = tf.boolean_mask(bbox, is_not_crowd), tf.boolean_mask(label, is_not_crowd)

        if self.magnitude >= 0:
            image, bbox = random_flip_left_right_with_bboxes(image, bbox)

        image, bbox = self.__random_crop_and_resize__(image, bbox)
        bbox, label = refine_bboxes_labels_single(bbox, label)

        if self.magnitude > 0:
            image.set_shape([*self.target_shape[:2], 3])
            image = self.randaug_wo_pos(image)

        should_pad = self.max_labels_per_image - tf.shape(bbox)[0]
        # label starts from 0 -> person, add 1 here, differs from padded values
        bbox, label = tf.cond(
            should_pad > 0,
            lambda: (tf.pad(bbox, [[0, should_pad], [0, 0]]), tf.pad(label + 1, [[0, should_pad]])),
            lambda: (bbox[: self.max_labels_per_image], label[: self.max_labels_per_image] + 1),
        )

        image = tf.cast(image, tf.float32)
        image.set_shape([*self.target_shape[:2], 3])
        return image, (bbox, label)


class PositionalRandAugmentWithBboxes:
    def __init__(self, magnitude=0, num_layers=2, max_labels_per_image=100, positional_augment_methods="rts", **randaug_kwargs):
        from keras_cv_attention_models.imagenet import augment

        self.pos_randaug = augment.PositionalRandAugment(
            num_layers=num_layers, magnitude=magnitude, positional_augment_methods=positional_augment_methods, **randaug_kwargs
        )
        self.max_labels_per_image = max_labels_per_image

    def __call_single__(self, inputs):
        image, bbox, label = inputs
        image, affine_matrix = self.pos_randaug(image)
        pick = tf.where(label != 0)
        bbox, label = tf.gather_nd(bbox, pick), tf.gather_nd(label, pick)
        bbox = bboxes_apply_affine(bbox, affine_matrix, input_shape=image.shape)
        bbox, label = refine_bboxes_labels_single(bbox, label)

        should_pad = self.max_labels_per_image - tf.shape(bbox)[0]
        bbox, label = tf.cond(
            should_pad > 0,
            lambda: (tf.pad(bbox, [[0, should_pad], [0, 0]]), tf.pad(label, [[0, should_pad]])),
            lambda: (bbox[: self.max_labels_per_image], label[: self.max_labels_per_image]),
        )
        return image, bbox, label

    def __call__(self, xx, yy):
        image, bbox, label = tf.map_fn(self.__call_single__, (xx, yy[0], yy[1]))
        return image, (bbox, label)


def to_one_hot_with_class_mark(anchor_bboxes_with_label, num_classes=80):
    # dest_boxes, anchor_classes = anchor_bboxes_with_label[:, :4], anchor_bboxes_with_label[:, -1]
    dest_boxes, anchor_classes = tf.split(anchor_bboxes_with_label, [4, 1], axis=-1)
    one_hot_labels = tf.one_hot(tf.cast(anchor_classes[..., 0], "int32") - 1, num_classes)  # [1, 81] -> [0, 80]
    # Mark iou < ignore_threshold as 0, ignore_threshold < iou < overlap_threshold as -1, iou > overlap_threshold as 1
    marks = tf.where(anchor_classes > 0, tf.ones_like(anchor_classes), anchor_classes)
    # marks = tf.expand_dims(marks, -1)
    one_hot_labels, marks = tf.cast(one_hot_labels, dest_boxes.dtype), tf.cast(marks, dest_boxes.dtype)
    return tf.concat([dest_boxes, one_hot_labels, marks], axis=-1)


def __bboxes_labels_batch_func__(bboxes, labels, anchors, empty_label, num_classes=80):
    bbox_labels = tf.concat([bboxes, tf.cast(tf.expand_dims(labels, -1), bboxes.dtype)], axis=-1)
    bbox_process = lambda xx: tf.cond(
        tf.reduce_any(xx[:, -1] > 0),  # If contains any valid bbox and label
        lambda: to_one_hot_with_class_mark(anchors_func.assign_anchor_classes_by_iou_with_bboxes(xx, anchors), num_classes),
        lambda: empty_label,
    )
    return tf.map_fn(bbox_process, bbox_labels)


def __yolor_bboxes_labels_batch_func__(bboxes, labels, anchor_ratios, feature_sizes, empty_label, num_classes=80):
    bbox_labels = tf.concat([bboxes, tf.cast(tf.expand_dims(labels, -1), bboxes.dtype)], axis=-1)
    bbox_process = lambda xx: tf.cond(
        tf.reduce_any(xx[:, -1] > 0),  # If contains any valid bbox and label
        lambda: to_one_hot_with_class_mark(anchors_func.yolor_assign_anchors(xx, anchor_ratios, feature_sizes), num_classes),
        lambda: empty_label,
    )
    return tf.map_fn(bbox_process, bbox_labels)


def detection_dataset_from_custom_json(data_path, with_info=False):
    import json

    with open(data_path, "r") as ff:
        aa = json.load(ff)

    test_key = "validation" if "validation" in aa else "test"
    train, test, info = aa["train"], aa[test_key], aa["info"]
    total_images, num_classes = len(train), info["num_classes"]
    objects_signature = {"bbox": tf.TensorSpec(shape=(None, 4), dtype=tf.float32), "label": tf.TensorSpec(shape=(None,), dtype=tf.int64)}
    output_signature = {"image": tf.TensorSpec(shape=(), dtype=tf.string), "objects": objects_signature}
    train_ds = tf.data.Dataset.from_generator(lambda: (ii for ii in train), output_signature=output_signature)
    test_ds = tf.data.Dataset.from_generator(lambda: (ii for ii in test), output_signature=output_signature)

    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    train_ds = train_ds.apply(tf.data.experimental.assert_cardinality(len(train))).with_options(options)
    test_ds = test_ds.apply(tf.data.experimental.assert_cardinality(len(test))).with_options(options)
    dataset = {"train": train_ds, test_key: test_ds}
    return (dataset, total_images, num_classes) if with_info else dataset


def init_dataset(
    data_name="coco/2017",  # dataset params
    input_shape=(256, 256),
    batch_size=64,
    buffer_size=1000,
    info_only=False,
    max_labels_per_image=100,
    anchors_mode="efficientdet",
    anchor_pyramid_levels=[3, 7],
    anchor_scale=4,  # For efficientdet anchors only
    aspect_ratios=(1, 2, 0.5),  # For efficientdet anchors only
    num_scales=3,  # For efficientdet anchors only
    rescale_mode="torch",  # rescale mode, ["tf", "torch"], or specific `(mean, std)` like `(128.0, 128.0)`
    resize_method="bilinear",  # ["bilinear", "bicubic"]
    resize_antialias=False,
    random_crop_mode=1.0,  # [augment] 0 for eval mode, (0, 1) for random crop, 1 for random largest crop, > 1 for random scale
    mosaic_mix_prob=0.0,  # [augment] 0 for disable, > 0 for enable
    color_augment_method="random_hsv",  # [augment] one of ["random_hsv", "autoaug", "randaug"], or totally custom one like `lambda image: image`
    positional_augment_methods="rts",  # [augment] Positional augment method besides scale, combine of r: rotate, t: transplate, s: shear, x: scale_x + scale_y
    magnitude=0,
    num_layers=2,
    seed=None,
    **augment_kwargs,  # Too many...
):
    import tensorflow_datasets as tfds

    is_tpu = True if len(tf.config.list_logical_devices("TPU")) > 0 else False  # Set True for try_gcs and drop_remainder

    if data_name.endswith(".json"):
        dataset, total_images, num_classes = detection_dataset_from_custom_json(data_name, with_info=True)
    else:
        dataset, info = tfds.load(data_name, with_info=True, try_gcs=is_tpu)
        num_classes = info.features["objects"]["label"].num_classes
        total_images = info.splits["train"].num_examples
    steps_per_epoch = int(tf.math.ceil(total_images / float(batch_size)))
    if info_only:
        return total_images, num_classes, steps_per_epoch

    AUTOTUNE = tf.data.AUTOTUNE

    train_process = RandomProcessImageWithBboxes(
        target_shape=input_shape,
        max_labels_per_image=max_labels_per_image,
        random_crop_mode=random_crop_mode,
        resize_method=resize_method,
        resize_antialias=resize_antialias,
        color_augment_method=color_augment_method,
        magnitude=magnitude,
        num_layers=num_layers,
        **augment_kwargs,
    )
    train_dataset = dataset["train"].shuffle(buffer_size, seed=seed).map(train_process).batch(batch_size, drop_remainder=is_tpu)
    # return train_dataset

    if mosaic_mix_prob > 0:
        mosaic_mix = lambda xx, yy: tf.cond(
            tf.random.uniform(()) > mosaic_mix_prob,
            lambda: (xx, yy),
            lambda: mosaic_mix_batch(xx, yy[0], yy[1]),
        )
        train_dataset = train_dataset.map(mosaic_mix, num_parallel_calls=AUTOTUNE)
    # return train_dataset

    if magnitude > 0 and positional_augment_methods is not None and len(positional_augment_methods) != 0:
        # Apply randaug rotate / shear / transform after mosaic mix
        max_labels_per_image = (max_labels_per_image * 4) if mosaic_mix_prob > 0 else max_labels_per_image
        pos_aug = PositionalRandAugmentWithBboxes(magnitude, num_layers, max_labels_per_image, positional_augment_methods, **augment_kwargs)
        print(">>>> positional augment methods:", pos_aug.pos_randaug.available_ops)
        train_dataset = train_dataset.map(pos_aug, num_parallel_calls=AUTOTUNE)

    if anchors_mode == anchors_func.ANCHOR_FREE_MODE:  # == "anchor_free"
        # Don't need anchors here, anchor assigning is after getting model predictions.
        bbox_process = lambda bb: to_one_hot_with_class_mark(tf.concat([bb[0], tf.cast(tf.expand_dims(bb[1], -1), bb[0].dtype)], axis=-1), num_classes)
    elif anchors_mode == anchors_func.YOLOR_MODE:  # == "yolor":
        anchor_ratios, feature_sizes = anchors_func.get_yolor_anchors(input_shape[:2], anchor_pyramid_levels, is_for_training=True)
        total_anchors = tf.cast(anchor_ratios.shape[1] * tf.reduce_sum(feature_sizes[:, 0] * feature_sizes[:, 1]), tf.int32)
        empty_label = tf.zeros([total_anchors, 4 + num_classes + 1])  # All 0
        bbox_process = lambda bb: __yolor_bboxes_labels_batch_func__(bb[0], bb[1], anchor_ratios, feature_sizes, empty_label, num_classes)
    else:
        # grid_zero_start = True if anchor_grid_zero_start == "auto" else anchor_grid_zero_start
        grid_zero_start = False  # Use this till meet some others new
        anchors = anchors_func.get_anchors(input_shape[:2], anchor_pyramid_levels, aspect_ratios, num_scales, anchor_scale, grid_zero_start)
        num_anchors = anchors.shape[0]
        empty_label = tf.zeros([num_anchors, 4 + num_classes + 1])  # All 0
        bbox_process = lambda bb: __bboxes_labels_batch_func__(bb[0], bb[1], anchors, empty_label, num_classes)

    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    rescaling = lambda xx: (xx - mean) / std
    train_dataset = train_dataset.map(lambda xx, yy: (rescaling(xx), bbox_process(yy)), num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.prefetch(buffer_size=AUTOTUNE)
    # return train_dataset

    """ Test dataset """
    test_dataset = dataset.get("validation", dataset.get("test", None))
    if test_dataset is not None:
        test_process = RandomProcessImageWithBboxes(target_shape=input_shape, resize_method=resize_method, resize_antialias=resize_antialias, magnitude=-1)
        test_dataset = test_dataset.map(test_process).batch(batch_size, drop_remainder=is_tpu).map(lambda xx, yy: (rescaling(xx), bbox_process(yy)))

    return train_dataset, test_dataset, total_images, num_classes, steps_per_epoch


""" Show """


def draw_bboxes(bboxes, ax=None):
    import matplotlib.pyplot as plt
    import numpy as np

    if ax is None:
        fig, ax = plt.subplots()
    bboxes = np.array(bboxes).astype("int32")
    for bb in bboxes:
        ax.plot(bb[[1, 1, 3, 3, 1]], bb[[0, 2, 2, 0, 0]])
    plt.show()
    return ax


def show_image_with_bboxes(
    image, bboxes, labels=None, confidences=None, is_bbox_width_first=False, ax=None, label_font_size=8, num_classes=80, indices_2_labels=None
):
    import matplotlib.pyplot as plt
    import numpy as np

    need_plt_show = False
    if ax is None:
        fig, ax = plt.subplots()
        need_plt_show = True
    ax.imshow(image)
    bboxes = np.array(bboxes)
    if is_bbox_width_first:
        bboxes = bboxes[:, [1, 0, 3, 2]]
    for id, bb in enumerate(bboxes):
        # bbox is [top, left, bottom, right]
        bb = [bb[0] * image.shape[0], bb[1] * image.shape[1], bb[2] * image.shape[0], bb[3] * image.shape[1]]
        bb = np.array(bb).astype("int32")
        ax.plot(bb[[1, 1, 3, 3, 1]], bb[[0, 2, 2, 0, 0]])

        if labels is not None:
            label = int(labels[id])
            if indices_2_labels is not None:
                label = indices_2_labels.get(label, indices_2_labels.get(str(label), "None"))
            elif num_classes == 90:
                label = COCO_90_LABEL_DICT[label]
            elif num_classes == 80:
                label = COCO_80_LABEL_DICT[label]

            if confidences is not None:
                label += ": {:.4f}".format(float(confidences[id]))
            color = ax.lines[-1].get_color()
            # ax.text(bb[1], bb[0] - 5, "label: {}, {}".format(label, COCO_80_LABEL_DICT[label]), color=color, fontsize=8)
            ax.text(bb[1], bb[0] - 5, label, color=color, fontsize=label_font_size)
    ax.set_axis_off()
    plt.tight_layout()
    if need_plt_show:
        plt.show()
    return ax


def show_batch_sample(
    dataset, rescale_mode="torch", rows=-1, label_font_size=8, base_size=3, anchors_mode="efficientdet", indices_2_labels=None, **anchor_kwargs
):
    import matplotlib.pyplot as plt
    from keras_cv_attention_models.visualizing import get_plot_cols_rows

    if isinstance(dataset, (list, tuple)):
        images, labels = dataset
    else:
        images, labels = dataset.as_numpy_iterator().next()
    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    images = (images * std + mean) / 255
    if anchors_mode == anchors_func.YOLOR_MODE:
        pyramid_levels = anchors_func.get_pyramid_levels_by_anchors(images.shape[1:-1], labels.shape[1])
        anchors = anchors_func.get_yolor_anchors(images.shape[1:-1], pyramid_levels=pyramid_levels, is_for_training=False)
    elif not anchors_mode == anchors_func.ANCHOR_FREE_MODE:
        pyramid_levels = anchors_func.get_pyramid_levels_by_anchors(images.shape[1:-1], labels.shape[1])
        anchors = anchors_func.get_anchors(images.shape[1:-1], pyramid_levels, **anchor_kwargs)

    cols, rows = get_plot_cols_rows(len(images), rows, ceil_mode=True)
    fig, axes = plt.subplots(rows, cols, figsize=(base_size * cols, base_size * rows))
    axes = axes.flatten()

    for ax, image, label in zip(axes, images, labels):
        if label.shape[-1] == 5:
            pick = label[:, -1] >= 0
            valid_preds = label[pick]
        else:
            pick = label[:, -1] == 1
            valid_preds = label[pick]
            valid_label = tf.cast(tf.argmax(valid_preds[:, 4:-1], axis=-1), valid_preds.dtype)
            valid_preds = tf.concat([valid_preds[:, :4], tf.expand_dims(valid_label, -1)], axis=-1)

        if anchors_mode == anchors_func.YOLOR_MODE:
            valid_anchors = anchors[pick]
            decoded_centers = (valid_preds[:, :2] + 0.5) * valid_anchors[:, 4:] + valid_anchors[:, :2]
            decoded_hw = valid_preds[:, 2:4] * valid_anchors[:, 4:]
            decoded_corner = anchors_func.center_yxhw_to_corners_nd(tf.concat([decoded_centers, decoded_hw], axis=-1))
            valid_preds = tf.concat([decoded_corner, valid_preds[:, -1:]], axis=-1)
        elif not anchors_mode == anchors_func.ANCHOR_FREE_MODE:
            valid_anchors = anchors[pick]
            valid_preds = anchors_func.decode_bboxes(valid_preds, valid_anchors)
        show_image_with_bboxes(image, valid_preds[:, :4], valid_preds[:, -1], ax=ax, label_font_size=label_font_size, indices_2_labels=indices_2_labels)
    fig.tight_layout()
    plt.show()
    return fig



from keras_cv_attention_models.aotnet import AotNet
from keras_cv_attention_models.download_and_load import reload_model_weights


PRETRAINED_DICT = {
    "resnext50": {"imagenet": "cf65d988c38ba0335c97a046288b91f4", "swsl": "f1cf0cc3c49bb50e6949c50fcce3db8f"},
    "resnext101": {"imagenet": "1e58c0ecc31184bd6bfe4d6b568f4325", "swsl": "c2fe8eefcf9a55e0254d2b13055a4cbc"},
    "resnext101w": {"imagenet": "9a1b92145aeb922695c29a0f02b52188", "swsl": "58b7cf4a72b03171f50ed19789b20f3d"},
    "resnext101w_64": {"imagenet": "51c81e014224bb731ebf64c3ed271a16"},
    "resnext50d": {"imagenet": "a7b2433b7bee7029fce11ba3fabf3fb9"},
}


def ResNeXt(num_blocks, input_shape=(224, 224, 3), pretrained="imagenet", strides=2, groups=32, **kwargs):
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]
    model = AotNet(num_blocks, input_shape=input_shape, strides=strides, groups=groups, **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


def ResNeXt50(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 3]
    hidden_channel_ratio = 0.5
    return ResNeXt(**locals(), model_name="resnext50", **kwargs)


def ResNeXt101(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 23, 3]
    hidden_channel_ratio = 0.5
    return ResNeXt(**locals(), model_name="resnext101", **kwargs)


def ResNeXt50D(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 3]
    hidden_channel_ratio = 0.5
    stem_type = "deep"
    shortcut_type = "avg"
    return ResNeXt(**locals(), model_name="resnext50d", **kwargs)


def ResNeXt101W(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 23, 3]
    hidden_channel_ratio = 1
    return ResNeXt(**locals(), model_name="resnext101w", **kwargs)


def ResNeXt101W_se(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained=None, **kwargs):
    # timm using an additional conv + bn before se_module
    num_blocks = [3, 4, 23, 3]
    hidden_channel_ratio = 1
    se_ratio = 0.25 / 4
    stem_type = "deep"
    return ResNeXt(**locals(), model_name="resnext101w", **kwargs)


def ResNeXt101W_64(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 23, 3]
    hidden_channel_ratio = 1
    groups = 64
    return ResNeXt(**locals(), model_name="resnext101w_64", **kwargs)



from keras_cv_attention_models.mobilenetv3_family.mobilenetv3 import MobileNetV3


def get_expanded_width_depth(width, depth):
    out_channels = [ii * width for ii in [16, 24, 40, 80, 112, 192, 320]]
    num_blocks = [int(round(ii * depth)) for ii in [1, 2, 2, 3, 3, 4, 1]]
    return out_channels, num_blocks


def TinyNet(
    num_blocks=[1, 2, 2, 3, 3, 4, 1],
    out_channels=[16, 24, 40, 80, 112, 192, 320],
    expands=[1, 6, 6, 6, 6, 6, 6],
    kernel_sizes=[3, 3, 5, 3, 5, 5, 3],
    strides=[1, 2, 2, 2, 1, 2, 1],
    activations="swish",
    stem_width=32,
    fix_stem=True,
    se_ratios=0.25,
    se_activation=None,  # None for same with activations
    use_expanded_se_ratio=False,
    se_divisor=1,
    output_num_features=1280,
    use_additional_output_conv=False,
    use_output_feature_bias=False,
    use_avg_pool_conv_output=False,
    model_name="tinynet",
    **kwargs,
):
    stem_feature_activation = activations
    kwargs.pop("kwargs", None)
    return MobileNetV3(**locals(), **kwargs)


def TinyNetA(input_shape=(192, 192, 3), num_classes=1000, activations="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, num_blocks = get_expanded_width_depth(1.0, 1.2)
    return TinyNet(**locals(), model_name="tinynet_a", **kwargs)


def TinyNetB(input_shape=(188, 188, 3), num_classes=1000, activations="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, num_blocks = get_expanded_width_depth(0.75, 1.1)
    return TinyNet(**locals(), model_name="tinynet_b", **kwargs)


def TinyNetC(input_shape=(184, 184, 3), num_classes=1000, activations="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, num_blocks = get_expanded_width_depth(0.54, 0.85)
    return TinyNet(**locals(), model_name="tinynet_c", **kwargs)


def TinyNetD(input_shape=(152, 152, 3), num_classes=1000, activations="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, num_blocks = get_expanded_width_depth(0.54, 0.695)
    return TinyNet(**locals(), model_name="tinynet_d", **kwargs)


def TinyNetE(input_shape=(106, 106, 3), num_classes=1000, activations="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, num_blocks = get_expanded_width_depth(0.51, 0.6)
    return TinyNet(**locals(), model_name="tinynet_e", **kwargs)



from keras_cv_attention_models import backend

from keras_cv_attention_models.coco import eval_func, anchors_func
from keras_cv_attention_models.coco.eval_func import DecodePredictions, COCOEvalCallback
from keras_cv_attention_models.coco.anchors_func import (
    get_anchors_mode_parameters,
    get_anchors,
    get_anchor_free_anchors,
    get_yolor_anchors,
    get_anchors_mode_by_anchors,
    get_pyramid_levels_by_anchors,
    decode_bboxes,
)

if backend.is_tensorflow_backend:
    from keras_cv_attention_models.coco import data, losses
    from keras_cv_attention_models.coco.data import (
        COCO_80_LABEL_DICT,
        COCO_90_LABEL_DICT,
        COCO_80_to_90_LABEL_DICT,
        aspect_aware_resize_and_crop_image,
        init_mean_std_by_rescale_mode,
        init_dataset,
        show_image_with_bboxes,
        show_batch_sample
    )

    data.init_dataset.__doc__ = """ Init dataset by name.
    Args:
      data_name: the registered dataset name from `tensorflow_datasets`.
      input_shape: input shape.
      batch_size: batch size.
      buffer_size: dataset shuffle buffer size.
      info_only: boolean value if returns dataset info only.
      max_labels_per_image: .
      anchors_mode: .
      anchor_pyramid_levels: .
      anchor_aspect_ratios: .
      anchor_num_scales: .
      anchor_scale: .
      anchor_scale: .
      cutmix_alpha: cutmix applying probability.
      rescale_mode: one of ["tf", "torch", "raw01", "raw"]. Detail in `data.init_mean_std_by_rescale_mode`. Or specific `(mean, std)` like `(128.0, 128.0)`.
      random_crop_mode: .
      mosaic_mix_prob: .
      resize_method: one of ["nearest", "bilinear", "bicubic"]. Resize method for `tf.image.resize`.
      resize_antialias: boolean value if using antialias for `tf.image.resize`.
      magnitude: randaug magnitude.
      num_layers: randaug num_layers.
      augment_kwargs: randaug kwargs. Too many to list them all.

    Returns: train_dataset, test_dataset, total_images, num_classes, steps_per_epoch
    """



from keras_cv_attention_models.levit.levit import (
    LeViT,
    LeViT128S,
    LeViT128,
    LeViT192,
    LeViT256,
    LeViT384,
    MultiHeadPositionalEmbedding,
    mhsa_with_multi_head_position_and_strides,
    mhsa_with_multi_head_position,
)


__head_doc__ = """
Keras implementation of [Github facebookresearch/LeViT](https://github.com/facebookresearch/LeViT).
Paper [PDF 2104.01136 LeViT: a Vision Transformer in ConvNet’s Clothing for Faster Inference](https://arxiv.org/pdf/2104.01136.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `hard_swish`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: top dropout rate if top layers is included. Default 0.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  use_distillation: Boolean value if output `distill_head`. Default `True`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

LeViT.__doc__ = __head_doc__ + """
Args:
  patch_channel: channel dimension output for `path_stem`.
  out_channels: output channels for each stack.
  num_heads: heads number for transformer blocks in each stack.
  depthes: number of block for each stack.
  key_dims: key dimension for transformer blocks in each stack.
  attn_ratios: `value` channel dimension expansion for transformer blocks in each stack.
  mlp_ratios: dimension expansion ration for `mlp_block` in each stack.
  strides: strides for each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                   | Params | FLOPs | Input | Top1 Acc |
  | ----------------------- | ------ | ----- | ----- | -------- |
  | LeViT128S, distillation | 7.8M   | 0.31G | 224   | 76.6     |
  | LeViT128, distillation  | 9.2M   | 0.41G | 224   | 78.6     |
  | LeViT192, distillation  | 11M    | 0.66G | 224   | 80.0     |
  | LeViT256, distillation  | 19M    | 1.13G | 224   | 81.6     |
  | LeViT384, distillation  | 39M    | 2.36G | 224   | 82.6     |
"""

LeViT128S.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

LeViT128.__doc__ = LeViT128S.__doc__
LeViT192.__doc__ = LeViT128S.__doc__
LeViT256.__doc__ = LeViT128S.__doc__
LeViT384.__doc__ = LeViT128S.__doc__

MultiHeadPositionalEmbedding.__doc__ = __head_doc__ + """
Multi Head Positional Embedding layer.

Positional embedding shape is `[height * width, num_heads]`.
input: `[batch, num_heads, query_blocks, key_blocks]`.
output: `[batch, num_heads, query_blocks, key_blocks] + positional_bias`.
conditions: key_height >= query_height, key_width >= query_width

Args:
  query_height: specify `height` for `query_blocks` if not square `query_height != query_width`.
  key_height: specify `height` for `key_blocks`.
      `-1` for auto calculating from `query_height` and `kk_blocks / qq_blocks` ratio. Works in most cases.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.MultiHeadPositionalEmbedding()
>>> print(f"{aa(tf.ones([1, 8, 16, 49])).shape = }")
# aa(tf.ones([1, 8, 16, 49])).shape = TensorShape([1, 8, 16, 49])

>>> print({ii.name:ii.numpy().shape for ii in aa.weights})
# {'multi_head_positional_embedding/positional_embedding:0': (49, 8)}

>>> plt.imshow(aa.bb_pos)
"""

mhsa_with_multi_head_position.__doc__ = __head_doc__ + """
Multi Head Self Attention with MultiHeadPositionalEmbedding.
Using additional `BatchNormalization` for `query / key / value`,
and adding `MultiHeadPositionalEmbedding` to `attention_scores`.

input: `[batch, height, width, channel]`.
output: `[batch, height, width, channel]`.

Args:
  inputs: Input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `-1` for `key_dim = input_channels // num_heads`.
  output_dim: The expected channel dimension of output. Default `-1` for same with input.
  attn_ratio: value channel dimension expansion.
  use_bn: boolean value if use BN layers for qkv and output.
  qkv_bias: boolean value if use bias for qkv layer.
  out_bias: boolean value if use bias for output layer.
  activation: activation for output, `None` to disable.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 12, 192])
>>> nn = attention_layers.mhsa_with_multi_head_position(inputs, output_dim=384, num_heads=4, key_dim=16)
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 12, 384])

>>> mm = keras.models.Model(inputs, nn)
>>> print({ii.name:ii.numpy().shape for ii in mm.weights})
# {'dense/kernel:0': (192, 192),
#  'batch_normalization/gamma:0': (192,),
#  'batch_normalization/beta:0': (192,),
#  'batch_normalization/moving_mean:0': (192,),
#  'batch_normalization/moving_variance:0': (192,),
#  'multi_head_positional_embedding/positional_embedding:0': (168, 4),
#  'dense_1/kernel:0': (64, 384),
#  'batch_normalization_1/gamma:0': (384,),
#  'batch_normalization_1/beta:0': (384,),
#  'batch_normalization_1/moving_mean:0': (384,),
#  'batch_normalization_1/moving_variance:0': (384,)}
"""

mhsa_with_multi_head_position_and_strides.__doc__ = __head_doc__ + """
Multi Head Self Attention with MultiHeadPositionalEmbedding and enabled strides on `query`.
Using additional `BatchNormalization` for `query / key / value`,
and adding `MultiHeadPositionalEmbedding` to `attention_scores`.
Also with a `strides` parameter which can further reduce calculation.

input: `[batch, height, width, channel]`.
output: `[batch, height, width, channel]`.

Args:
  inputs: Input tensor.
  output_dim: The expected channel dimension of output.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `-1` for `key_dim = input_channels // num_heads`.
  output_dim: The expected channel dimension of output. Default `-1` for same with input.
  attn_ratio: value channel dimension expansion.
  strides: query strides on height and width dimension.
  use_bn: boolean value if use BN layers for qkv and output.
  qkv_bias: boolean value if use bias for qkv layer.
  out_bias: boolean value if use bias for output layer.
  activation: activation for output, `None` to disable.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 12, 192])
>>> nn = attention_layers.mhsa_with_multi_head_position_and_strides(inputs, output_dim=384, num_heads=4, key_dim=16, strides=2)
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 7, 6, 384])

>>> mm = keras.models.Model(inputs, nn)
>>> print({ii.name:ii.numpy().shape for ii in mm.weights})
# {'dense_4/kernel:0': (192, 192),
#  'dense_3/kernel:0': (192, 64),
#  'batch_normalization_4/gamma:0': (192,),
#  'batch_normalization_4/beta:0': (192,),
#  'batch_normalization_4/moving_mean:0': (192,),
#  'batch_normalization_4/moving_variance:0': (192,),
#  'batch_normalization_3/gamma:0': (64,),
#  'batch_normalization_3/beta:0': (64,),
#  'batch_normalization_3/moving_mean:0': (64,),
#  'batch_normalization_3/moving_variance:0': (64,),
#  'multi_head_positional_embedding_1/positional_embedding:0': (168, 4),
#  'dense_5/kernel:0': (128, 384),
#  'batch_normalization_5/gamma:0': (384,),
#  'batch_normalization_5/beta:0': (384,),
#  'batch_normalization_5/moving_mean:0': (384,),
#  'batch_normalization_5/moving_variance:0': (384,)}
"""



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers
from keras_cv_attention_models.attention_layers import activation_by_name, add_pre_post_process
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "resmlp12": {"imagenet": "de6531fb461bcf52c25d3c36aa515583"},
    "resmlp24": {"imagenet": "f8127be7f8ba564fc59552c0cf6f3401"},
    "resmlp36": {"imagenet": "d0d3e6b09d7e975aaf46ff777c1fd73e"},
    "resmlp_b24": {"imagenet": "d7808ef59c06d2f1975ffddd28be82de", "imagenet22k": "8d3ae1abdac60b21ed1f2840b656b6bf"},
}


@backend.register_keras_serializable(package="resmlp")
class ChannelAffine(layers.Layer):
    def __init__(self, use_bias=True, weight_init_value=1, axis=-1, **kwargs):
        super().__init__(**kwargs)
        self.use_bias, self.weight_init_value, self.axis = use_bias, weight_init_value, axis
        self.ww_init = initializers.Constant(weight_init_value) if weight_init_value != 1 else "ones"
        self.bb_init = "zeros"
        self.supports_masking = False

    def build(self, input_shape):
        if self.axis == -1 or self.axis == len(input_shape) - 1:
            ww_shape = (input_shape[-1],)
        else:
            ww_shape = [1] * len(input_shape)
            axis = self.axis if isinstance(self.axis, (list, tuple)) else [self.axis]
            for ii in axis:
                ww_shape[ii] = input_shape[ii]
            ww_shape = ww_shape[1:]  # Exclude batch dimension

        self.ww = self.add_weight(name="weight", shape=ww_shape, initializer=self.ww_init, trainable=True)
        if self.use_bias:
            self.bb = self.add_weight(name="bias", shape=ww_shape, initializer=self.bb_init, trainable=True)
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        return inputs * self.ww + self.bb if self.use_bias else inputs * self.ww

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        if backend.image_data_format() != "channels_last" and self.axis == 1:
            weights = [np.squeeze(ii) for ii in weights]
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        if backend.image_data_format() != "channels_last" and self.axis == 1:
            weights = [np.reshape(ii, self.ww.shape) for ii in weights]
        return self.set_weights(weights)

    def get_config(self):
        config = super().get_config()
        config.update({"use_bias": self.use_bias, "weight_init_value": self.weight_init_value, "axis": self.axis})
        return config


# NOT using
def channel_affine(inputs, use_bias=True, weight_init_value=1, name=""):
    ww_init = initializers.Constant(weight_init_value) if weight_init_value != 1 else "ones"
    nn = functional.expand_dims(inputs, 1)
    nn = layers.DepthwiseConv2D(1, depthwise_initializer=ww_init, use_bias=use_bias, name=name)(nn)
    return functional.squeeze(nn, 1)


def res_mlp_block(inputs, channels_mlp_dim, drop_rate=0, activation="gelu", name=None):
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[-1] if backend.image_data_format() == "channels_last" else inputs.shape[1]

    nn = ChannelAffine(use_bias=True, axis=channel_axis, name=name + "norm_1")(inputs)
    nn = layers.Permute((2, 1), name=name + "permute_1")(nn) if backend.image_data_format() == "channels_last" else nn
    nn = layers.Dense(nn.shape[-1], name=name + "token_mixing")(nn)
    nn = layers.Permute((2, 1), name=name + "permute_2")(nn) if backend.image_data_format() == "channels_last" else nn
    nn = ChannelAffine(use_bias=False, axis=channel_axis, name=name + "gamma_1")(nn)
    if drop_rate > 0:
        nn = layers.Dropout(drop_rate, noise_shape=(None, 1, 1), name=name + "token_drop")(nn)
    token_out = layers.Add(name=name + "add_1")([inputs, nn])

    nn = ChannelAffine(use_bias=True, axis=channel_axis, name=name + "norm_2")(token_out)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((2, 1), name=name and name + "permute_3")(nn)
    nn = layers.Dense(channels_mlp_dim, name=name + "channel_mixing_1")(nn)
    nn = activation_by_name(nn, activation, name=name + activation)
    nn = layers.Dense(input_channel, name=name + "channel_mixing_2")(nn)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((2, 1), name=name and name + "permute_4")(nn)
    channel_out = ChannelAffine(use_bias=False, axis=channel_axis, name=name + "gamma_2")(nn)
    if drop_rate > 0:
        channel_out = layers.Dropout(drop_rate, noise_shape=(None, 1, 1), name=name + "channel_drop")(channel_out)
    # print(f">>>> {inputs.shape = }, {channel_out.shape = }, {token_out.shape = }")
    nn = layers.Add(name=name + "output")([channel_out, token_out])
    return nn


def ResMLP(
    num_blocks,
    patch_size,
    stem_width,
    channels_mlp_dim,
    input_shape=(224, 224, 3),
    num_classes=0,
    activation="gelu",
    sam_rho=0,
    dropout=0,
    drop_connect_rate=0,
    classifier_activation="softmax",
    pretrained="imagenet",
    model_name="resmlp",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    nn = layers.Conv2D(stem_width, kernel_size=patch_size, strides=patch_size, padding="valid", name="stem")(inputs)
    new_shape = [nn.shape[1] * nn.shape[2], stem_width] if backend.image_data_format() == "channels_last" else [stem_width, nn.shape[2] * nn.shape[3]]
    nn = layers.Reshape(new_shape)(nn)

    drop_connect_s, drop_connect_e = drop_connect_rate if isinstance(drop_connect_rate, (list, tuple)) else [drop_connect_rate, drop_connect_rate]
    for ii in range(num_blocks):
        name = "{}_{}_".format("ResMlpBlock", str(ii + 1))
        block_drop_rate = drop_connect_s + (drop_connect_e - drop_connect_s) * ii / num_blocks
        nn = res_mlp_block(nn, channels_mlp_dim=channels_mlp_dim, drop_rate=block_drop_rate, activation=activation, name=name)
    nn = ChannelAffine(axis=-1 if backend.image_data_format() == "channels_last" else 1, name="pre_head_norm")(nn)

    if num_classes > 0:
        # nn = tf.reduce_mean(nn, axis=1)
        nn = layers.GlobalAveragePooling1D()(nn)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    if sam_rho != 0:
        from keras_cv_attention_models.model_surgery import SAMModel

        model = SAMModel(inputs, nn, name=model_name)
    else:
        model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="mlp_family", pretrained=pretrained)
    return model


BLOCK_CONFIGS = {
    "12": {
        "num_blocks": 12,
        "patch_size": 16,
        "stem_width": 384,
        "channels_mlp_dim": 384 * 4,
    },
    "24": {
        "num_blocks": 24,
        "patch_size": 16,
        "stem_width": 384,
        "channels_mlp_dim": 384 * 4,
    },
    "36": {
        "num_blocks": 36,
        "patch_size": 16,
        "stem_width": 384,
        "channels_mlp_dim": 384 * 4,
    },
    "b24": {
        "num_blocks": 24,
        "patch_size": 8,
        "stem_width": 768,
        "channels_mlp_dim": 768 * 4,
    },
}


def ResMLP12(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return ResMLP(**BLOCK_CONFIGS["12"], **locals(), model_name="resmlp12", **kwargs)


def ResMLP24(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return ResMLP(**BLOCK_CONFIGS["24"], **locals(), model_name="resmlp24", **kwargs)


def ResMLP36(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return ResMLP(**BLOCK_CONFIGS["36"], **locals(), model_name="resmlp36", **kwargs)


def ResMLP_B24(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return ResMLP(**BLOCK_CONFIGS["b24"], **locals(), model_name="resmlp_b24", **kwargs)



"""
Creates a EfficientNetV2 Model as defined in: Mingxing Tan, Quoc V. Le. (2021). arXiv preprint arXiv:2104.00298.
EfficientNetV2: Smaller Models and Faster Training.
"""
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, is_channels_last
from keras_cv_attention_models.attention_layers import (
    batchnorm_with_activation,
    conv2d_no_bias,
    drop_block,
    global_context_module,
    make_divisible,
    output_block,
    se_module,
    add_pre_post_process,
)

TF_BATCH_NORM_EPSILON = 0.001
TORCH_BATCH_NORM_EPSILON = 1e-5


FILE_HASH_DICT = {
    "v2-b0": {"21k-ft1k": "4e4da4eb629897e4d6271e131039fe75", "21k": "5dbb4252df24b931e74cdd94d150f25a", "imagenet": "9abdc43cb00f4cb06a8bdae881f412d6"},
    "v2-b1": {"21k-ft1k": "5f1aee82209f4f0f20bd24460270564e", "21k": "a50ae65b50ceff7f5283be2f4506d2c2", "imagenet": "5d4223b59ff268828d5112a1630e234e"},
    "v2-b2": {"21k-ft1k": "ec384b84441ddf6419938d1e5a0cbef2", "21k": "9f718a8bbb7b63c5313916c5e504790d", "imagenet": "1814bc08d4bb7a5e0ed3ccfe1cf18650"},
    "v2-b3": {"21k-ft1k": "4a27827b0b2df508bed31ae231003bb1", "21k": "ade5bdbbdf1d54c4561aa41511525855", "imagenet": "cda85b8494c7ec5a68dffb335a254bab"},
    "v2-l": {"21k-ft1k": "30327edcf1390d10e9a0de42a2d731e3", "21k": "7970f913eec1b4918e007c8580726412", "imagenet": "2b65f5789f4d2f1bf66ecd6d9c5c2d46"},
    "v2-m": {"21k-ft1k": "0c236c3020e3857de1e5f2939abd0cc6", "21k": "3923c286366b2a5137f39d1e5b14e202", "imagenet": "ac3fd0ff91b35d18d1df8f1895efe1d5"},
    "v2-s": {"21k-ft1k": "93046a0d601da46bfce9d4ca14224c83", "21k": "10b05d878b64f796ab984a5316a4a1c3", "imagenet": "3b91df2c50c7a56071cca428d53b8c0d"},
    "v2-t": {"imagenet": "4a0ff9cb396665734d7ca590fa29681b"},
    "v2-t-gc": {"imagenet": "653fc06396f9503dff61aa17c40b2443"},
    "v2-xl": {"21k-ft1k": "9aaa2bd3c9495b23357bc6593eee5bce", "21k": "c97de2770f55701f788644336181e8ee"},
    "v1-b0": {"noisy_student": "d125a518737c601f8595937219243432", "imagenet": "cc7d08887de9df8082da44ce40761986"},
    "v1-b1": {"noisy_student": "8f44bff58fc5ef99baa3f163b3f5c5e8", "imagenet": "a967f7be55a0125c898d650502c0cfd0"},
    "v1-b2": {"noisy_student": "b4ffed8b9262df4facc5e20557983ef8", "imagenet": "6c8d1d3699275c7d1867d08e219e00a7"},
    "v1-b3": {"noisy_student": "9d696365378a1ebf987d0e46a9d26ddd", "imagenet": "d78edb3dc7007721eda781c04bd4af62"},
    "v1-b4": {"noisy_student": "a0f61b977544493e6926186463d26294", "imagenet": "4c83aa5c86d58746a56675565d4f2051"},
    "v1-b5": {"noisy_student": "c3b6eb3f1f7a1e9de6d9a93e474455b1", "imagenet": "0bda50943b8e8d0fadcbad82c17c40f5"},
    "v1-b6": {"noisy_student": "20dd18b0df60cd7c0387c8af47bd96f8", "imagenet": "da13735af8209f675d7d7d03a54bfa27"},
    "v1-b7": {"noisy_student": "7f6f6dd4e8105e32432607ad28cfad0f", "imagenet": "d9c22b5b030d1e4f4c3a96dbf5f21ce6"},
    "v1-l2": {"noisy_student": "5fedc721febfca4b08b03d1f18a4a3ca"},
}


def inverted_residual_block(
    inputs,
    output_channel,
    stride=1,
    expand=4,
    shortcut=False,
    kernel_size=3,
    drop_rate=0,
    se_ratio=0,
    is_fused=False,
    is_torch_mode=False,
    se_activation=None,  # None for same with activation
    se_divisor=1,  # 8 for mobilenetv3
    se_limit_round_down=0.9,  # 0.95 for fbnet
    use_global_context_instead_of_se=False,
    use_last_bn_zero_gamma=False,
    activation="swish",
    name=None,
):
    input_channel = inputs.shape[-1 if is_channels_last() else 1]
    bn_eps = TORCH_BATCH_NORM_EPSILON if is_torch_mode else TF_BATCH_NORM_EPSILON
    hidden_channel = make_divisible(input_channel * expand, 8)

    if is_fused and expand != 1:
        nn = conv2d_no_bias(inputs, hidden_channel, 3, stride, padding="same", use_torch_padding=is_torch_mode, name=name and name + "sortcut_")
        nn = batchnorm_with_activation(nn, activation=activation, epsilon=bn_eps, name=name and name + "sortcut_")
    elif expand != 1:
        nn = conv2d_no_bias(inputs, hidden_channel, 1, strides=1, padding="valid", use_torch_padding=is_torch_mode, name=name and name + "sortcut_")
        nn = batchnorm_with_activation(nn, activation=activation, epsilon=bn_eps, name=name and name + "sortcut_")
    else:
        nn = inputs

    if not is_fused:
        if is_torch_mode and backend.is_tensorflow_backend and kernel_size // 2 > 0:
            nn = layers.ZeroPadding2D(padding=kernel_size // 2, name=name and name + "pad")(nn)
            padding = "VALID"
        elif is_torch_mode and kernel_size // 2 > 0:
            padding = kernel_size // 2
        else:
            padding = "SAME"
        nn = layers.DepthwiseConv2D(kernel_size, padding=padding, strides=stride, use_bias=False, name=name and name + "MB_dw_")(nn)
        nn = batchnorm_with_activation(nn, activation=activation, epsilon=bn_eps, name=name and name + "MB_dw_")

    if se_ratio > 0:
        se_activation = activation if se_activation is None else se_activation
        se_ratio = se_ratio / expand
        if use_global_context_instead_of_se:
            nn = global_context_module(nn, use_attn=True, ratio=se_ratio, divisor=1, activation=se_activation, use_bias=True, name=name and name + "gc_")
        else:
            nn = se_module(nn, se_ratio, divisor=se_divisor, limit_round_down=se_limit_round_down, activation=se_activation, name=name and name + "se_")

    # pw-linear
    if is_fused and expand == 1:
        nn = conv2d_no_bias(nn, output_channel, 3, strides=stride, padding="same", use_torch_padding=is_torch_mode, name=name and name + "fu_")
        nn = batchnorm_with_activation(nn, activation=activation, zero_gamma=use_last_bn_zero_gamma, epsilon=bn_eps, name=name and name + "fu_")
    else:
        nn = conv2d_no_bias(nn, output_channel, 1, strides=1, padding="valid", use_torch_padding=is_torch_mode, name=name and name + "MB_pw_")
        nn = batchnorm_with_activation(nn, activation=None, zero_gamma=use_last_bn_zero_gamma, epsilon=bn_eps, name=name and name + "MB_pw_")

    if shortcut:
        nn = drop_block(nn, drop_rate, name=name and name + "drop")
        return layers.Add(name=name and name + "output")([inputs, nn])
    else:
        return layers.Activation("linear", name=name and name + "output")(nn)  # Identity, Just need a name here


def EfficientNetV2(
    expands=[1, 4, 4, 4, 6, 6],
    out_channels=[16, 32, 48, 96, 112, 192],
    depthes=[1, 2, 2, 3, 5, 8],
    strides=[1, 2, 2, 2, 1, 2],
    se_ratios=[0, 0, 0, 0.25, 0.25, 0.25],
    is_fused="auto",  # True if se_ratio == 0 else False
    first_conv_filter=32,
    output_conv_filter=1280,
    kernel_sizes=3,
    input_shape=(None, None, 3),
    num_classes=1000,
    dropout=0.2,
    first_strides=2,
    is_torch_mode=False,
    use_global_context_instead_of_se=False,
    drop_connect_rate=0,
    activation="swish",
    classifier_activation="softmax",
    include_preprocessing=False,
    pretrained="imagenet",
    model_name="EfficientNetV2",
    rescale_mode="torch",
    kwargs=None,
):
    # "torch" for all V1 models
    # for V2 models, "21k" pretrained are all "tf", "imagenet" pretrained "bx" models are all "torch", ["s", "m", "l", "xl"] are "tf"
    rescale_mode = "tf" if pretrained is not None and pretrained.startswith("imagenet21k") else rescale_mode

    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(shape=input_shape)
    bn_eps = TORCH_BATCH_NORM_EPSILON if is_torch_mode else TF_BATCH_NORM_EPSILON

    if include_preprocessing and rescale_mode == "torch":
        channel_axis = 1 if backend.image_data_format() == "channels_first" else -1
        Normalization = layers.Normalization if hasattr(layers, "Normalization") else layers.experimental.preprocessing.Normalization
        mean = np.array([0.485, 0.456, 0.406], dtype="float32") * 255.0
        std = (np.array([0.229, 0.224, 0.225], dtype="float32") * 255.0) ** 2
        nn = Normalization(mean=mean, variance=std, axis=channel_axis)(inputs)
    elif include_preprocessing and rescale_mode == "tf":
        Rescaling = layers.Rescaling if hasattr(layers, "Rescaling") else layers.experimental.preprocessing.Rescaling
        nn = Rescaling(scale=1.0 / 128.0, offset=-1)(inputs)
    else:
        nn = inputs
    stem_width = make_divisible(first_conv_filter, 8)
    nn = conv2d_no_bias(nn, stem_width, 3, strides=first_strides, padding="same", use_torch_padding=is_torch_mode, name="stem_")
    nn = batchnorm_with_activation(nn, activation=activation, epsilon=bn_eps, name="stem_")

    blocks_kwargs = {  # common for all blocks
        "is_torch_mode": is_torch_mode,
        "use_global_context_instead_of_se": use_global_context_instead_of_se,
    }

    pre_out = stem_width
    global_block_id = 0
    total_blocks = sum(depthes)
    kernel_sizes = kernel_sizes if isinstance(kernel_sizes, (list, tuple)) else ([kernel_sizes] * len(depthes))
    for id, (expand, out_channel, depth, stride, se_ratio, kernel_size) in enumerate(zip(expands, out_channels, depthes, strides, se_ratios, kernel_sizes)):
        out = make_divisible(out_channel, 8)
        if is_fused == "auto":
            cur_is_fused = True if se_ratio == 0 else False
        else:
            cur_is_fused = is_fused[id] if isinstance(is_fused, (list, tuple)) else is_fused
        for block_id in range(depth):
            name = "stack_{}_block{}_".format(id, block_id)
            stride = stride if block_id == 0 else 1
            shortcut = True if out == pre_out and stride == 1 else False
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = inverted_residual_block(
                nn, out, stride, expand, shortcut, kernel_size, block_drop_rate, se_ratio, cur_is_fused, **blocks_kwargs, activation=activation, name=name
            )
            pre_out = out
            global_block_id += 1

    if output_conv_filter > 0:
        output_conv_filter = make_divisible(output_conv_filter, 8)
        nn = conv2d_no_bias(nn, output_conv_filter, 1, strides=1, padding="valid", use_torch_padding=is_torch_mode, name="post_")
        nn = batchnorm_with_activation(nn, activation=activation, epsilon=bn_eps, name="post_")
    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)

    model = models.Model(inputs=inputs, outputs=nn, name=model_name)
    add_pre_post_process(model, rescale_mode="raw" if include_preprocessing else rescale_mode)
    reload_model_weights(model, pretrained)
    return model


def reload_model_weights(model, pretrained="imagenet"):
    import os

    if pretrained is None:
        return
    if isinstance(pretrained, str) and pretrained.endswith(".h5"):
        print(">>>> Load pretrained from:", pretrained)
        model.load_weights(pretrained, by_name=True, skip_mismatch=True)
        return

    pretrained_dd = {"imagenet": "imagenet", "imagenet21k": "21k", "imagenet21k-ft1k": "21k-ft1k", "noisy_student": "noisy_student"}
    if not pretrained in pretrained_dd:
        print(">>>> No pretrained available, model will be randomly initialized")
        return
    pre_tt = pretrained_dd[pretrained]
    model_type = model.name.split("_")[-1]
    if model_type not in FILE_HASH_DICT or pre_tt not in FILE_HASH_DICT[model_type]:
        print(">>>> No pretrained available, model will be randomly initialized")
        return

    if model_type.startswith("v1"):
        pre_url = "https://github.com/leondgarse/keras_efficientnet_v2/releases/download/effnetv1_pretrained/efficientnet{}-{}.h5"
    else:
        pre_url = "https://github.com/leondgarse/keras_efficientnet_v2/releases/download/effnetv2_pretrained/efficientnet{}-{}.h5"
    url = pre_url.format(model_type, pre_tt)
    file_name = os.path.basename(url)
    file_hash = FILE_HASH_DICT[model_type][pre_tt]

    try:
        pretrained_model = backend.get_file(file_name, url, cache_subdir="models/efficientnetv2", file_hash=file_hash)
    except:
        print("[Error] will not load weights, url not found or download failed:", url)
        return
    else:
        print(">>>> Load pretrained from:", pretrained_model)
        model.load_weights(pretrained_model, by_name=True, skip_mismatch=True)


def EfficientNetV2B0(input_shape=(224, 224, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.0, depth 1.0
    out_channels = [16, 32, 48, 96, 112, 192]
    depthes = [1, 2, 2, 3, 5, 8]
    first_conv_filter = kwargs.pop("first_conv_filter", 32)
    output_conv_filter = kwargs.pop("output_conv_filter", 1280)
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-b0", **kwargs)


def EfficientNetV2B1(input_shape=(240, 240, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.0, depth 1.1
    out_channels = [16, 32, 48, 96, 112, 192]
    depthes = [2, 3, 3, 4, 6, 9]
    first_conv_filter = kwargs.pop("first_conv_filter", 32)
    output_conv_filter = kwargs.pop("output_conv_filter", 1280)
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-b1", **kwargs)


def EfficientNetV2B2(input_shape=(260, 260, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.1, depth 1.2
    out_channels = [16, 32, 56, 104, 120, 208]
    depthes = [2, 3, 3, 4, 6, 10]
    first_conv_filter = kwargs.pop("first_conv_filter", 32)
    output_conv_filter = kwargs.pop("output_conv_filter", 1408)
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-b2", **kwargs)


def EfficientNetV2B3(input_shape=(300, 300, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.2, depth 1.4
    out_channels = [16, 40, 56, 112, 136, 232]
    depthes = [2, 3, 3, 5, 7, 12]
    first_conv_filter = kwargs.pop("first_conv_filter", 40)
    output_conv_filter = kwargs.pop("output_conv_filter", 1536)
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-b3", **kwargs)


def EfficientNetV2T(input_shape=(288, 288, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.4 * 0.8, depth 1.8 * 0.9, from timm
    is_torch_mode = True
    out_channels = [24, 40, 48, 104, 128, 208]
    depthes = [2, 4, 4, 6, 9, 14]
    first_conv_filter = kwargs.pop("first_conv_filter", 24)
    output_conv_filter = kwargs.pop("output_conv_filter", 1024)
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-t", **kwargs)


def EfficientNetV2T_GC(input_shape=(288, 288, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.4 * 0.8, depth 1.8 * 0.9, from timm
    is_torch_mode = True
    use_global_context_instead_of_se = True
    out_channels = [24, 40, 48, 104, 128, 208]
    depthes = [2, 4, 4, 6, 9, 14]
    first_conv_filter = kwargs.pop("first_conv_filter", 24)
    output_conv_filter = kwargs.pop("output_conv_filter", 1024)
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-t-gc", **kwargs)


def EfficientNetV2S(input_shape=(384, 384, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.4, depth 1.8
    out_channels = [24, 48, 64, 128, 160, 256]
    depthes = [2, 4, 4, 6, 9, 15]
    first_conv_filter = kwargs.pop("first_conv_filter", 24)
    output_conv_filter = kwargs.pop("output_conv_filter", 1280)
    rescale_mode = kwargs.pop("rescale_mode", "tf")
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-s", **kwargs)


def EfficientNetV2M(input_shape=(480, 480, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.6, depth 2.2
    out_channels = [24, 48, 80, 160, 176, 304, 512]
    depthes = [3, 5, 5, 7, 14, 18, 5]
    expands = [1, 4, 4, 4, 6, 6, 6]
    strides = [1, 2, 2, 2, 1, 2, 1]
    se_ratios = [0, 0, 0, 0.25, 0.25, 0.25, 0.25]
    first_conv_filter = kwargs.pop("first_conv_filter", 24)
    output_conv_filter = kwargs.pop("output_conv_filter", 1280)
    rescale_mode = kwargs.pop("rescale_mode", "tf")
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-m", **kwargs)


def EfficientNetV2L(input_shape=(480, 480, 3), num_classes=1000, dropout=0.4, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # width 1.6, depth 2.2
    out_channels = [32, 64, 96, 192, 224, 384, 640]
    depthes = [4, 7, 7, 10, 19, 25, 7]
    expands = [1, 4, 4, 4, 6, 6, 6]
    strides = [1, 2, 2, 2, 1, 2, 1]
    se_ratios = [0, 0, 0, 0.25, 0.25, 0.25, 0.25]
    first_conv_filter = kwargs.pop("first_conv_filter", 32)
    output_conv_filter = kwargs.pop("output_conv_filter", 1280)
    rescale_mode = kwargs.pop("rescale_mode", "tf")
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-l", **kwargs)


def EfficientNetV2XL(input_shape=(512, 512, 3), num_classes=1000, dropout=0.4, classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    out_channels = [32, 64, 96, 192, 256, 512, 640]
    depthes = [4, 8, 8, 16, 24, 32, 8]
    expands = [1, 4, 4, 4, 6, 6, 6]
    strides = [1, 2, 2, 2, 1, 2, 1]
    se_ratios = [0, 0, 0, 0.25, 0.25, 0.25, 0.25]
    first_conv_filter = kwargs.pop("first_conv_filter", 32)
    output_conv_filter = kwargs.pop("output_conv_filter", 1280)
    rescale_mode = kwargs.pop("rescale_mode", "tf")
    return EfficientNetV2(**locals(), model_name="efficientnet_v2-xl", **kwargs)



import numpy as np


def get_plot_cols_rows(total, rows=-1, ceil_mode=False):
    if rows == -1 and total < 8:
        rows = 1  # for total in [1, 7], plot 1 row only
    elif rows == -1:
        rr = int(np.floor(np.sqrt(total)))
        for ii in range(1, rr + 1)[::-1]:
            if total % ii == 0:
                rows = ii
                break
    if ceil_mode:
        cols = int(np.ceil(total / rows))
    else:
        cols = total // rows
    return cols, rows


def put_text_on_image(image, text, coord=(5, 5), color=(255, 0, 0)):
    from PIL import Image
    from PIL import ImageDraw

    # from PIL import ImageFont

    image = image * 255 if image.max() < 2 else image
    img = Image.fromarray(image.astype("uint8"))
    draw = ImageDraw.Draw(img)
    draw.text(coord, str(text), color)
    return np.array(img)


def stack_and_plot_images(images, texts=None, margin=5, margin_value=0, rows=-1, ax=None, base_size=3):
    """Stack and plot a list of images. Returns ax, stacked_images"""
    import matplotlib.pyplot as plt

    cols, rows = get_plot_cols_rows(len(images), rows, ceil_mode=True)
    if cols * rows > len(images):
        padded = cols * rows - len(images)
        images += [np.zeros_like(images[-1]) + 255 for ii in range(padded)]
    # images = images[: rows * cols]
    # print(">>>> rows:", rows, ", cols:", cols, ", total:", len(images))

    if texts is not None:
        images = [put_text_on_image(imm, itt) for imm, itt in zip(images, texts)] + list(images[len(texts) :])
        images = np.array(images)

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(base_size * cols, base_size * rows))

    if margin > 0:
        ww_margin = np.zeros_like(images[0][:, :margin]) + margin_value
        ww_margined_images = [np.hstack([ii, ww_margin]) for ii in images]
        hstacked_images = [np.hstack(ww_margined_images[ii : ii + cols]) for ii in range(0, len(ww_margined_images), cols)]

        hh_margin = np.zeros_like(hstacked_images[0][:margin]) + margin_value
        hh_margined_images = [np.vstack([ii, hh_margin]) for ii in hstacked_images]
        vstacked_images = np.vstack(hh_margined_images)

        stacked_images = vstacked_images[:-margin, :-margin]
    else:
        stacked_images = np.vstack([np.hstack(images[ii * cols : (rr + 1) * cols]) for ii in range(rows)])

    ax.imshow(stacked_images)
    ax.set_axis_off()
    ax.grid(False)
    plt.tight_layout()
    # plt.show()
    return ax, stacked_images


""" tensorboard_parallel_coordinates_plot """


def tensorboard_parallel_coordinates_plot(dataframe, metrics_name, metrics_display_name=None, skip_columns=[], log_dir="logs/hparam_tuning"):
    import os
    import pandas as pd
    import tensorflow as tf
    from tensorboard.plugins.hparams import api as hp

    skip_columns = skip_columns + [metrics_name]
    to_hp_discrete = lambda column: hp.HParam(column, hp.Discrete(np.unique(dataframe[column].values).tolist()))
    hp_params_dict = {column: to_hp_discrete(column) for column in dataframe.columns if column not in skip_columns}

    if dataframe[metrics_name].values.dtype == "object":  # Not numeric
        import json

        metrics_map = {ii: id for id, ii in enumerate(np.unique(dataframe[metrics_name]))}
        description = json.dumps(metrics_map)
    else:
        metrics_map, description = None, None

    METRICS = metrics_name if metrics_display_name is None else metrics_display_name
    with tf.summary.create_file_writer(log_dir).as_default():
        metrics = [hp.Metric(METRICS, display_name=METRICS, description=description)]
        hp.hparams_config(hparams=list(hp_params_dict.values()), metrics=metrics)

    for id in dataframe.index:
        log = dataframe.iloc[id]
        hparams = {hp_unit: log[column] for column, hp_unit in hp_params_dict.items()}
        print({hp_unit.name: hparams[hp_unit] for hp_unit in hparams})
        run_dir = os.path.join(log_dir, "run-%d" % id)
        with tf.summary.create_file_writer(run_dir).as_default():
            hp.hparams(hparams)  # record the values used in this trial
            metric_item = log[metrics_name] if metrics_map is None else metrics_map[log[metrics_name]]
            tf.summary.scalar(METRICS, metric_item, step=1)

    print()
    if metrics_map is not None:
        print(">>>> metrics_map:", metrics_map)
    print(">>>> Start tensorboard by: ! tensorboard --logdir {}".format(log_dir))
    print(">>>> Then select `HPARAMS` -> `PARALLEL COORDINATES VIEW`")



from keras_cv_attention_models.caformer.caformer import (
    CAFormer,
    CAFormerS18,
    CAFormerS36,
    CAFormerM36,
    CAFormerB36,
    ConvFormerS18,
    ConvFormerS36,
    ConvFormerM36,
    ConvFormerB36,
)

__head_doc__ = """
Keras implementation of [Github sail-sg/metaformer](https://github.com/sail-sg/metaformer).
Paper [PDF 2210.13452 MetaFormer Baselines for Vision](https://arxiv.org/pdf/2210.13452.pdf).

`CAFormer` is using 2 transformer stacks by `block_types=["conv", "conv", "transform", "transform"]`.
`ConvFormer` is all conv blocks by `block_types=["conv", "conv", "conv", "conv"]`.
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  layer_scales: int or list of int, indicates layer scale init value for each stack. Default `[0, 0, 1e-6, 1e-6]`, 0 for not using.
      [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
  residual_scales: int or list of int, indicates residual short branch scale init value for each stack.
      Default `[0, 0, 1, 1]`, 0 for not using.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet' or 'imagenet21k-ft1k' (pre-training on ImageNet21k and fine-tuned ImageNet).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

CAFormer.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  block_types: block types for each stack,
      - `conv` or any `c` / `C` starts word, means `mlp_block_with_depthwise_conv` block.
      - `transfrom` or any `t` / `T` starts word, means `multi_head_self_attention` block.
      value could be in format like `"cctt"` or `"CCTT"` or `["conv", "conv", "transfrom", "transform"]`.
  head_dim: heads number for transformer block.
  mlp_ratios: int or list value indicates expand ratio for mlp blocks hidden channel in each stack.
  head_filter: numer of feature block channels before output block. 0 for not using.
  head_filter_activation: feature block activation. None for same with activation.
  num_attn_low_heads: int or list value indicates attention heads number for `attention_low_frequency_mixer` blocks in each stack.
  pool_sizes: int or list value indicates attention blocks key_value downsample rate in each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model              | Params | FLOPs | Input | Top1 Acc |
  | ------------------ | ------ | ----- | ----- | -------- |
  | CAFormerS18        | 26M    | 4.1G  | 224   | 83.6     |
  | - imagenet21k-ft1k | 26M    | 4.1G  | 224   | 84.1     |
  |                    | 26M    | 13.4G | 384   | 85.0     |
  | - imagenet21k-ft1k | 26M    | 13.4G | 384   | 85.4     |
  | CAFormerS36        | 39M    | 8.0G  | 224   | 84.5     |
  | - imagenet21k-ft1k | 39M    | 8.0G  | 224   | 85.8     |
  |                    | 39M    | 26.0G | 384   | 85.7     |
  | - imagenet21k-ft1k | 39M    | 26.0G | 384   | 86.9     |
  | CAFormerM36        | 56M    | 13.2G | 224   | 85.2     |
  | - imagenet21k-ft1k | 56M    | 13.2G | 224   | 86.6     |
  |                    | 56M    | 42.0G | 384   | 86.2     |
  | - imagenet21k-ft1k | 56M    | 42.0G | 384   | 87.5     |
  | CAFormerB36        | 99M    | 23.2G | 224   | 85.5     |
  | - imagenet21k-ft1k | 99M    | 23.2G | 224   | 87.4     |
  |                    | 99M    | 72.2G | 384   | 86.4     |
  | - imagenet21k-ft1k | 99M    | 72.2G | 384   | 88.1     |

  | Model              | Params | FLOPs | Input | Top1 Acc |
  | ------------------ | ------ | ----- | ----- | -------- |
  | ConvFormerS18      | 27M    | 3.9G  | 224   | 83.0     |
  | - imagenet21k-ft1k | 27M    | 3.9G  | 224   | 83.7     |
  |                    | 27M    | 11.6G | 384   | 84.4     |
  | - imagenet21k-ft1k | 27M    | 11.6G | 384   | 85.0     |
  | ConvFormerS36      | 40M    | 7.6G  | 224   | 84.1     |
  | - imagenet21k-ft1k | 40M    | 7.6G  | 224   | 85.4     |
  |                    | 40M    | 22.4G | 384   | 85.4     |
  | - imagenet21k-ft1k | 40M    | 22.4G | 384   | 86.4     |
  | ConvFormerM36      | 57M    | 12.8G | 224   | 84.5     |
  | - imagenet21k-ft1k | 57M    | 12.8G | 224   | 86.1     |
  |                    | 57M    | 37.7G | 384   | 85.6     |
  | - imagenet21k-ft1k | 57M    | 37.7G | 384   | 86.9     |
  | ConvFormerB36      | 100M   | 22.6G | 224   | 84.8     |
  | - imagenet21k-ft1k | 100M   | 22.6G | 224   | 87.0     |
  |                    | 100M   | 66.5G | 384   | 85.7     |
  | - imagenet21k-ft1k | 100M   | 66.5G | 384   | 87.6     |
"""

CAFormerS18.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

CAFormerS36.__doc__ = CAFormerS18.__doc__
CAFormerM36.__doc__ = CAFormerS18.__doc__
CAFormerB36.__doc__ = CAFormerS18.__doc__
ConvFormerS18.__doc__ = CAFormerS18.__doc__
ConvFormerS36.__doc__ = CAFormerS18.__doc__
ConvFormerM36.__doc__ = CAFormerS18.__doc__
ConvFormerB36.__doc__ = CAFormerS18.__doc__
